{*****************************************************************}
{                                                                 }
{             ScriptReporter 2.1 example unit                     }
{                                                                 }
{  Function                 ACC_XOR(Str1, Str2)                   }
{                           accumulative xor                      }
{  --------                                                       }
{  Call from script example  Result := ACC_XOR(query->field);     }
{  ------------------------                                       }
{                                                                 }
{  Author: Vlad Ilushchenko                                       }
{                                                                 }
{*****************************************************************}

unit ExAcc;

interface

uses Reporter, RepExpr, RepKrnl;

type
  TAccumulativeExample = class(TRepUnaryExpression)
  private
    // value on the current step of reporter
    FOurValue: Integer;
    // value on the previous step of reporter
    FOurPreviousValue: Integer;
  protected
    function GetAsVariant: Variant; override;
    //
    // you should override this method for using your
    // function in the following constructions:
    //
    // Result := ACC_XOR(query->field_1) when Check(query->field_2) reset to 0;
    //
    function GetOldValue: Variant; override;
    procedure SetAsVariant(Value: Variant); override;
    procedure Prepare; override;
    procedure Turn; override;
  public
    constructor Create(RepEngine: TCustomRepEngine;
                       Expr, WhenClause, ResetClause: TRepToken;
                       AName: String; Pos: TRepPosition); override;
  end;

implementation

constructor TAccumulativeExample.Create;
begin
  inherited Create(RepEngine, Expr, WhenClause, ResetClause, AName, Pos);

  // Reporter will call to Turn procedure on each step,
  // regardless to using this expression on the current step
  Turnable := True;
end;

function TAccumulativeExample.GetOldValue: Variant;
begin
  Result := FOurPreviousValue;
end;

function TAccumulativeExample.GetAsVariant: Variant;
begin
  Result := FOurValue;
end;

procedure TAccumulativeExample.SetAsVariant(Value: Variant);
begin
  FOurValue := Value;
end;

procedure TAccumulativeExample.Prepare;
begin
  // SelfPrepared is set to True if you function have been prepared
  if SelfPrepared then exit;

  // you MUST call to inherited method for proper proceeding
  inherited Prepare;

  // initialize our variable
  FOurValue := 0;

  SelfPrepared := True;
end;

procedure TAccumulativeExample.Turn;
begin
  // this method should define the behaviour of the function
  FOurPreviousValue := FOurValue;

  // prevent from invalid variant type conversion at the end of the report
  if Arg.Addr.Value <> Null then
    FOurValue := FOurValue xor Arg.Addr.Value;
end;

initialization

  AddPlugIn('ACC_XOR', [akRight], TAccumulativeExample, False);
//           ^         ^          ^                     ^
//           |         |          |                     will not treated as separate token
//           |         |          class reference
//           |         argument decription
//           plug-in name

end.
