{**
*
*         Script Reporter Component library
*         Debugger using example
*          Vlad Ilushchenko 1998-99
*
*
**}
unit DebuggerForm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, RepDebugger, Buttons, Oracle, RepViewForm, StdCtrls, ScriptRep, RepKrnl,
  Menus;

type
  TfmDebugger = class(TForm)
    Panel1: TPanel;
    btnOpen: TSpeedButton;
    btnSaveAs: TSpeedButton;
    btnRun: TSpeedButton;
    RepDebugger: TRepDebugger;
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    btnSave: TSpeedButton;
    btnClose: TSpeedButton;
    cbWaitForTerminate: TCheckBox;
    PopupMenu1: TPopupMenu;
    Openfile1: TMenuItem;
    Savefile1: TMenuItem;
    N1: TMenuItem;
    Run1: TMenuItem;
    N2: TMenuItem;
    Apply1: TMenuItem;
    Close1: TMenuItem;
    Label1: TLabel;
    Label2: TLabel;
    procedure btnRunClick(Sender: TObject);
    procedure btnOpenClick(Sender: TObject);
    procedure btnSaveAsClick(Sender: TObject);
    procedure btnSaveClick(Sender: TObject);
    procedure btnCloseClick(Sender: TObject);
    procedure RepDebuggerAfterRun(Sender: TRepDebugger;
      Reporter: TScriptRep);
    procedure RepDebuggerBeforeRun(Sender: TRepDebugger;
      Reporter: TScriptRep);
    procedure RepDebuggerError(Sender: TRepDebugger; Reporter: TScriptRep);
  private
    { Private declarations }
    FFileName: String;
    procedure LoadFile(const FileName: string);
    procedure Save;
    procedure SaveTo(const FileName: String);

    procedure RepTerminated(Sender: TCustomRepEngine; TerminateType: TTerminateType);
    procedure LineProcessed(Sender: TObject; Line: Integer);
  public
    { Public declarations }
    procedure OpenDebugger(const FileName: String;  Session: TOracleSession);
  end;

var
  fmDebugger: TfmDebugger;

implementation

{$R *.DFM}

procedure TfmDebugger.LoadFile(const FileName: string);
begin
  RepDebugger.Lines.LoadFromFile(FileName);
  FFileName := FileName;
  Caption := FFileName;
end;

procedure TfmDebugger.Save;
begin
  RepDebugger.Lines.SaveToFile(FFileName);
end;

procedure TfmDebugger.SaveTo(const FileName: String);
begin
  RepDebugger.Lines.SaveToFile(FileName);
  FFileName := FileName;
  Caption := FileName;
end;

procedure TfmDebugger.OpenDebugger(const FileName: String; Session: TOracleSession);
begin
  fmRepView := TfmRepView.Create(Self);
  try
    RepDebugger.Session := Session;
    RepDebugger.Reporter.OnTerminate := RepTerminated;
    RepDebugger.Reporter.OnLineProcessed := LineProcessed;
    LoadFile(FileName);
    ActiveControl := RepDebugger;
    if ShowModal = mrOk then Save;
  finally
    fmRepView.Free;
  end;
end;

procedure TfmDebugger.RepTerminated(Sender: TCustomRepEngine;
                                    TerminateType: TTerminateType);
begin
  fmRepView.btnStop.Enabled := False;
end;

procedure TfmDebugger.btnRunClick(Sender: TObject);
begin
  fmRepView.btnStop.Enabled := True;
  if cbWaitForTerminate.Checked then
    RepDebugger.RunScript(wtTerminate)
  else
    RepDebugger.RunScript(wtFirstPage);
end;

procedure TfmDebugger.btnOpenClick(Sender: TObject);
begin
  if OpenDialog.Execute then LoadFile(OpenDialog.FileName);
end;

procedure TfmDebugger.btnSaveAsClick(Sender: TObject);
begin
  SaveTo(FFileName);
end;

procedure TfmDebugger.btnSaveClick(Sender: TObject);
begin
  ModalResult := mrOk;
end;

procedure TfmDebugger.btnCloseClick(Sender: TObject);
begin
  ModalResult := mrCancel;
end;

procedure TfmDebugger.RepDebuggerAfterRun(Sender: TRepDebugger;
  Reporter: TScriptRep);
begin
  Screen.Cursor := crDefault;
  fmRepView.FileView.FileName := Reporter.OutputFile;
  fmRepView.Caption := Reporter.Title;
  fmRepView.ShowModal;
  Label1.Caption := 'Done!';
end;

procedure TfmDebugger.RepDebuggerBeforeRun(Sender: TRepDebugger;
  Reporter: TScriptRep);
begin
  Screen.Cursor := crHourGlass;
  Label1.Caption := 'Running...'
end;

procedure TfmDebugger.RepDebuggerError(Sender: TRepDebugger;
  Reporter: TScriptRep);
begin
  Screen.Cursor := crDefault;
  Label1.Caption := 'Terminated';
end;

procedure TfmDebugger.LineProcessed;
begin
  Label1.Caption := Format('Compiling:        %d lines processed', [Line]);
  Label1.Update;
end;


end.
