{**
*
*         Script Reporter Component library
*         File view component
*          Vlad Ilushchenko 1998-99
*
*
**}

unit FileView;

interface

uses Classes, SysUtils, ObjList, Controls, Windows, Graphics, Messages,
     PrintSetup, Dialogs;

const
  InitialBufLength = 400;

type
  ERangeError = class(Exception) end;

  TCustomView = class;

  TViewMode = (vmPages, vmClear);

  TViewOption = (voRowSelect,
                 voCaret,
                 voTabs,
                 voPersistentBlocks,
                 voPrintToDOSDevice);

  TLineReadingState = (lrsInitializing,
                       lrsWindowSwap,
                       lrsWindowsPrinting,
                       lrsDOSPrinting,
                       lrsBlockWriting,
                       lrsDOSBlockWriting);

  TViewOptions = set of TViewOption;
  TViewFindOption = (vfoIgnoreCase,
                     vfoWholeWord,
                     vfoForward,
                     vfoEntireScope);
  TViewFindoptions = set of TViewFindoption;

  TPrintAction = (paAbort, paIgnoreCurrent, paIgnoreAll);

  TViewPrintNotifyEvent = procedure(Sender: TCustomView; var
                                    PrintAction: TPrintAction) of object;

  TStringEncoder = procedure(Sender: TCustomView; var Str: String;
                             State: TLineReadingState) of Object;

  TTextPos = record
    Line: Integer;
    Col:  Integer;
  end;

  TMarkType = (mtVertical, mtHorizontal, mtNone, mtAny);

  TPageInfo = class(TPersistent)
  private
    FLeftMargin: Integer;
    FRightMargin: Integer;
    FTopMargin: Integer;
    FBottomMargin: Integer;
    FHeight: Integer;
    FWidth: Integer;
    FPageSpacing: Integer;
    FLineGap: Integer;
  public
    procedure Assign(Source: TPersistent); override;
  published
    property LeftMargin: Integer read FLeftMargin write FLeftMargin;
    property RightMargin: Integer read FRightMargin write FRightMargin;
    property TopMargin: Integer read FTopMargin write FTopMargin;
    property BottomMargin: Integer read FBottomMargin write FBottomMargin;
    property Height: Integer read FHeight write FHeight;
    property Width: Integer read FWidth write FWidth;
    property PageSpacing: Integer read FPageSpacing write FPageSpacing;
    property LineGap: Integer read FLineGap write FLineGap;
  end;

  TColorSpec = class(TPersistent)
  private
    FTextColor: TColor;
    FBackColor: TColor;
    FView: TCustomView;

    procedure SetTextColor(Value: TColor);
    procedure SetBackColor(Value: TColor);
  public
    constructor Create(View: TCustomView);
    procedure Assign(Source: TPersistent); override;
  published
    property TextColor: TColor read FTextColor write SetTextColor;
    property BackColor: TColor read FBackColor write SetBackColor;
  end;

  TLineProcessor = procedure (S: TStrings;
                              LocalIndex: Integer;
                              GlobalIndex: Integer;
                              R: TRect) of object;
  TCaret = class
  private
    FAllocated: Boolean;
    FHeight: Longint;
    FParent: TWinControl;
    FPos: TPoint;
    FVisible: Boolean;
  protected
    procedure Recreate;
    procedure SetHeight(Value: Longint);
    procedure SetVisible(Value: Boolean);
    procedure SetPos(Value: TPoint);
  public
    constructor Create(Parent: TWinControl; Height: Longint);
    destructor Destroy; override;
    procedure Hide;
    procedure Show;
    procedure Remove;
    property Height: Longint read FHeight write SetHeight;
    property Pos: TPoint read FPos write SetPos;
    property Visible: Boolean read FVisible write SetVisible;
  end;

  TCustomView = class(TWinControl)
  private
    FActiveDC: HDC;
    FPageInfo: TPageInfo;
    FPageInfoMm: TPageInfo;
    FUpperBuffer: TStrings;
    FLowerBuffer: TStrings;
    FFileName:    String;
    FStream:      TStream;
    FLineIndex:   Integer;
    FGlobalLine:  Integer;
    FViewMode:    TViewMode;
    FBuffer:      PChar;
    FBrush:       TBrush;
    FActiveBrush: TBrush;
    FActiveLineShift: Integer;
    FActivePosShift: Integer;
    FLineHeight:  Integer;
    FStartPos: Integer;
    FMidPos:   Integer;
    FEndPos:   Integer;
    FHorzShift: Integer;
    FCaret: TCaret;
    FTempPageInfo: TPageInfo;
    FMouseCaptured: Boolean;
    FFileSize: Integer;
    FLinesCount: Integer;
    FViewOptions: TViewOptions;
    FSilentMove: Boolean;
    FLastFoundLine: Integer;
    FLastFoundCol: Integer;
    FLinesPerPage: Integer;
    FDOSPrintDevice: String;
    FPrintFontSize: Integer;
    FPrintingType: TPrintingType;
    FMousePoint: TPoint;
    FTimerID: Integer;
    FOnPageOutOfWidth: TViewPrintNotifyEvent;
    //
    FBlockColor: TColorSpec;
    FSelectedColor: TColorSpec;
    FSelectedBlockColor: TColorSpec;
    //
    FMarking: Boolean;
    FMarkType:  TMarkType;
    FMarkBegin: TTextPos;
    FMarkEnd:   TTextPos;
    FMarkStartPoint: TTextPos;
    FGlobalMarkOffset: Integer;
    //
    FStringEncoder: TStringEncoder;
    //
    FClipboardHandle: THandle;
    //-------parameters for line processors----------
    FDC: HDC;
    FRect: TRect;
    FVisibleLinesCount: Integer;
    FPoint: TPoint;
    FNewLine: Integer;
    FWasSwap: Boolean;
    //-----------------------------------------------
    function ReadLineBack(var S: String; ReadingState: TLineReadingState): Boolean;
    function ReadLineForward(var S: String; ReadingState: TLineReadingState): Boolean;
    procedure SwapBuffers;
    function UpdateText(NeededShift: Integer): Integer;
    procedure RedrawView(DC: HDC);
    procedure InitialFillBuffers;
    procedure InvalidateView(Rect: TRect);
    procedure RecalcLineSize;
    function ActiveDC: HDC;
    procedure FreeActiveDC;
    //----------- line processors -----------------------
    procedure DrawLine(S: TStrings;
                       LocalIndex: Integer;
                       GlobalIndex: Integer;
                       R: TRect);
    procedure InvalidateLine(S: TStrings;
                             LocalIndex: Integer;
                             GlobalIndex: Integer;
                             R: TRect);
    procedure Counter(S: TStrings;
                      LocalIndex: Integer;
                      GlobalIndex: Integer;
                      R: TRect);
    procedure FindLine(S: TStrings;
                      LocalIndex: Integer;
                      GlobalIndex: Integer;
                      R: TRect);

    //----------------------------------------------------
    procedure ScrollArea(Dx, Dy: Integer);
    procedure SetLineIndex(Value: Integer);
    procedure SetPageInfoMm(Value: TPageInfo);
    procedure UpdateActiveLine(HideIt: Boolean);
    procedure UpdateCaret;
    procedure SetLineByPoint(Pt: TPoint);
    procedure CountLinesAtFile;
    function  CalcPosShift(S: String; Shift: Integer): Integer;
    function  CalcRealShift(LineShift: Integer; Shift: Integer): Cardinal;
    function  CalcLeftWordOffset(S: String; FromShift: Integer; WordNum: Integer): Integer;
    function  CalcRightWordOffset(S: String; FromShift: Integer; WordNum: Integer): Integer;
    procedure SetViewOptions(Value: TViewOptions);
    procedure FreeClipboardHandle;
    function GetGlobalOffset(GlobalLine: Integer): Integer;
    procedure ColorChanged;
    procedure RecalcMargins;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure WMPaint(var Msg: TWMPaint); message WM_PAINT;
    procedure CMFontChanged(var Msg: TMessage); message CM_FONTCHANGED;
    procedure WMGetDlgCode(var Msg: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure CMShowingChanged(var Msg: TMessage); message CM_SHOWINGCHANGED;
    procedure KeyDown(var Key: Word; ShiftState: TShiftState); override;
    procedure KeyUp(var Key: Word; ShiftState: TShiftState); override;
    procedure WMLButtonDown(var Msg: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonUp(var Msg: TWMLButtonUp); message WM_LBUTTONUP;
    procedure WMTimer(var Msg: TWMTimer); message WM_TIMER;
    procedure WMMouseMove(var Msg: TWMMouseMove); message WM_MOUSEMOVE;
    procedure CMDesignHitTest(var Msg: TCMDesignHitTest); message CM_DESIGNHITTEST;
    procedure WMSetFocus(var Msg: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Msg: TWMKillFocus); message WM_KILLFOCUS;
    procedure CMColorChanged(var Msg: TMessage); message CM_COLORCHANGED;
    procedure CMCtl3DChanged(var Msg: TMessage); message CM_CTL3DCHANGED;
    procedure WMEraseBckgnd(var Msg: TMessage); message WM_ERASEBKGND;
    procedure WMDestroyClipboard(var Msg: TMessage); message WM_DESTROYCLIPBOARD;
    procedure WMDestroy(var Msg: TMessage); message WM_DESTROY;
    procedure WMVScroll(var Msg: TWMVScroll); message WM_VSCROLL;
    procedure WMHScroll(var Msg: TWMHScroll); message WM_HSCROLL;
    procedure WMCopy(var Msg: TWMCopy); message WM_COPY;
    procedure WMSize(var Msg: TWMSize); message WM_SIZE;
    function EnumVisibleLines(LineProcessor: TLineProcessor): Integer;

    function GetVisibleLinesCount: Integer;
    procedure TextLineUp;
    procedure TextLineDown;
    procedure ScrollText(Shift: Integer);

    procedure InternalMovePageUp;
    procedure InternalMovePageDown;
    procedure InternalMoveLineDown;
    procedure InternalMoveLineUp;
    procedure InternalMoveTop;
    procedure InternalMoveBottom;
    procedure InternalMoveLineHome;
    procedure InternalMoveLineEnd;
    procedure InternalMoveCaret(Count: Integer; MoveBehind: Boolean);
    procedure InternalMoveWord(Count: Integer);

    procedure MovePageUp;
    procedure MovePageDown;
    procedure MoveLineDown;
    procedure MoveLineUp;
    procedure MoveTop;
    procedure MoveBottom;
    procedure MoveLineHome;
    procedure MoveLineEnd;
    procedure MoveCaretLeft;
    procedure MoveCaretRight;
    procedure MoveWordLeft;
    procedure MoveWordRight;

    procedure StartMarking(MarkType: TMarkType);
    procedure UpdateMarking;
    procedure FinishMarking(MarkType: TMarkType);
    procedure CancelMarking;
    procedure CheckMarking;

    procedure InternalPrint(Title: String; StartPage, EndPage: Integer; IgnorePageBreaks: Boolean);
    procedure DOSPrint;
    procedure DOSPrintBlock(IgnorePageBreaks: Boolean);
    procedure DOSInternalPrint(StartPage, EndPage: Integer; IgnorePageBreaks: Boolean);
    procedure CopyToClipboard;
    property LineIndex: Integer read FLineIndex write SetLineIndex;
  public
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;

    {Closes the opened file}
    procedure CloseFile;

    {Opens the given file. The AFileName will kept opened unil you call to
     CloseFile method, or the component will be destroyed}
    procedure OpenFile(AFileName: String);

    {Saves file as...}
    procedure SaveAs(AFileName: String);

    {This function allows to you to check whether some text
     have been matched or not}
    function SomethingMatched: Boolean;

    {Finds the given text in the file. By default this function searches the
     text beckwards(e.g. from the current position of the caret to the top
     of the file) and it'll be case sensitive. Find options are:

     vfoEntireScope - force searching in the entire file
     vfoForward     - force searching to the end of file
     vfoIgnoreCase  - (no comments)
     vfoWholeWord   - IGNORED BY CURRENT VERSION (Sorry, i had no time to implement it)}
    function FindText(Text: String; FindOptions: TViewFindOptions): Boolean;

    {Writes matched block to stream}
    procedure WriteBlockToStream(S: TStream);

    {Saves the matched block to file}
    procedure SaveBlock(FileName: String);

    {Printing routines}

    {Printing concepts:
       the component:
       1. allows to print both to matrix and graphical devices
       2. allows to print the selected page or pages range
       3. supports page break symbol '#12' both for matrix and graphical devices
       4. allows to print blocks.

       By default the component will print to the Canvas of currently selected
       printer. If you want to print to the matrix printer (or port) you must
       include voPrintToDOSDevice flag in ViewOptions.
    }

    {Prints selected block to printer.
     Title            - printer job title
     IgnorePageBreaks - in case of TRUE all lines wich are contains '#12' symbol
                        will be ignored}
    procedure PrintBlock(Title: String; IgnorePageBreaks: Boolean);

    {Prints entire file
     Title - printer job title}
    procedure Print(Title: String; IgnorePageBreaks: Boolean);

    {Prints selected range of pages.
     Title    - printer job title
     FromPage - first page number
     ToPage   - Last page number

     This method assumes that page numbers are starts from 0.
     If you want to print to the end of file, you must set ToPage argument to -1
     }
    procedure PrintRange(Title: String; FromPage, ToPage: Integer; IgnorePageBreaks: Boolean);

    {Prints set of pages.
     PrintString - MS Word like print string, for example '1,15, 23-48, 3'}
    procedure PrintFromString(Title: String; PrintString: String; IgnorePageBreaks: Boolean);

    property PageInfo: TPageInfo read FPageInfo write FPageInfo;
    property PageInfoMm: TPageInfo read FPageInfoMm write SetPageInfoMm;
    property FileName: String read FFileName write OpenFile;
    property ViewOptions: TViewOptions read FViewOptions write SetViewOptions;
    property SelectedColor: TColorSpec read FSelectedColor write FSelectedColor;
    property BlockColor: TColorSpec read FBlockColor write FBlockColor;
    property SelectedBlockColor: TColorSpec read FSelectedBlockColor write FSelectedBlockColor;
    property OnConvertString: TStringEncoder read FStringEncoder write FStringEncoder;
    property LinesPerPage: Integer read FLinesPerPage write FLinesPerPage;
    property DOSPrintDevice: String read FDOSPrintDevice write FDOSPrintDevice;
    property PrintFontSize: Integer read FPrintFontSize write FPrintFontSize;
    property PrintingType: TPrintingType read FPrintingType write FPrintingType;
    property OnPageOutOfWidth: TViewPrintNotifyEvent read FOnPageOutOfWidth write FOnPageOutOfWidth;
  end;

  TFileView = class(TCustomView)
  published
    property Align;
    property PageInfoMm;
    property FileName;
    property Font;
    property ParentFont;
    property TabStop;
    property TabOrder;
    property ViewOptions;
    property Color;
    property SelectedColor;
    property BlockColor;
    property SelectedBlockColor;

    property OnClick;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseUp;
    property OnMouseMove;
    property OnKeyDown;
    property OnKeyUp;
    property OnKeyPress;
    property OnDblClick;
    property OnDragOver;
    property OnDragDrop;
    property OnStartDrag;
    property OnConvertString;
    property OnPageOutOfWidth;
    property LinesPerPage;
    property DOSPrintDevice;
    property PrintFontSize;
    property PrintingType;
  end;

procedure Register;

function TextPos(Line: Integer; Col: Integer): TTextPos;

implementation

{$R *.DCR}

uses DsgnIntf, StringTools, Clipbrd, Printers, Forms;

const
  CRLF = #13#10;
  MaxStringLen = 2048;

type
  TFileNamePropertyEditor = class(TStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure Edit; override;
  end;

procedure Register;
begin
  RegisterComponents('ISD', [TFileView]);
  RegisterPropertyEditor(TypeInfo(String), TFileView, 'FileName',
    TFileNamePropertyEditor);
end;


{Misc. procedures}

function TextPos(Line: Integer; Col: Integer): TTextPos;
begin
  Result.Line := Line;
  Result.Col := Col;
end;

function LessThen(P1, P2: TTextPos): Boolean;
begin
  Result := (P1.Line < P2.Line) or
            ((P1.Line = P2.Line) and (P1.Col < P2.Col));
end;

function LessOrEquel(P1, P2: TTextPos): Boolean;
begin
  Result := (P1.Line < P2.Line) or
            ((P1.Line = P2.Line) and (P1.Col <= P2.Col));
end;

function GreaterThen(P1, P2: TTextPos): Boolean;
begin
  Result := (P1.Line > P2.Line) or
            ((P1.Line = P2.Line) and (P1.Col > P2.Col));
end;

function GreaterOrEquel(P1, P2: TTextPos): Boolean;
begin
  Result := (P1.Line > P2.Line) or
            ((P1.Line = P2.Line) and (P1.Col >= P2.Col));
end;

function Equel(P1, P2: TTextPos): Boolean;
begin
  Result := (P1.Line = P2.Line) and (P1.Col = P2.Col);
end;


{TFileName editor}

function TFileNamePropertyEditor.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog];
end;

procedure TFileNamePropertyEditor.Edit;
var
  OpenDialog: TOpenDialog;
begin
  OpenDialog := TOpenDialog.Create(nil);
  try
    OpenDialog.Filter := 'Text files|*.txt|All files|*.*';
    OpenDialog.Options := [];
    OpenDialog.FileName := TCustomView(GetComponent(0)).FileName;
    if OpenDialog.Execute then
    begin
      TCustomView(GetComponent(0)).FileName := OpenDialog.FileName;
      Designer.Modified;
    end;

  finally
    OpenDialog.Free;
  end;
end;

{TPageInfo}

procedure TPageInfo.Assign(Source: TPersistent);
begin
  if Source = nil then exit;

  with Source do
  begin
    Self.FLeftMargin := FLeftMargin;
    Self.FRightMargin := FRightMargin;
    Self.FTopMargin := FTopMargin;
    Self.FBottomMargin := FBottomMargin;
    Self.FHeight := FHeight;
    Self.FWidth := FWidth;
    Self.FPageSpacing := FPageSpacing;
    Self.FLineGap := FLineGap;
  end;
end;

{TColorSpec}

constructor TColorSpec.Create(View: TCustomView);
begin
  FView := View;
  inherited Create;
end;

procedure TColorSpec.Assign(Source: TPersistent);
begin
  with TColorSpec(Source) do
  begin
    Self.FTextColor := TextColor;
    Self.FBackColor := BackColor;
    FView.ColorChanged;
  end;
end;

procedure TColorSpec.SetTextColor(Value: TColor);
begin
  FTextColor := Value;
  FView.ColorChanged;
end;

procedure TColorSpec.SetBackColor(Value: TColor);
begin
  FBackColor := Value;
  FView.ColorChanged;
end;



{TCaret}

constructor TCaret.Create(Parent: TWinControl; Height: Longint);
begin
  if CreateCaret(Parent.Handle, 0, 2, Height) then
  begin
    Self.FHeight := Height;
    Self.FParent := Parent;
    Self.FVisible := False;
    Self.FAllocated := True;
  end
  else
    raise Exception.Create(Format('Can''t create system caret. Error code %d', [GetLastError]));
end;

destructor TCaret.Destroy;
begin
  Remove;
end;

procedure TCaret.Hide;
begin
  if not FAllocated then Recreate;
  if FVisible then HideCaret(Self.FParent.Handle);
  FVisible := False
end;

procedure TCaret.Recreate;
begin
  Remove;
  FAllocated := True;
  if not CreateCaret(FParent.Handle, 0, 2 , FHeight) then
    raise Exception.Create(Format('Can''t create system caret. Error code %d', [GetLastError]));
  if FVisible then Show;
end;

procedure TCaret.Remove;
begin
  if FAllocated then
  begin
    if FVisible then Hide;
    DestroyCaret;
  end;
  FAllocated := False;
end;

procedure TCaret.SetHeight(Value: Longint);
begin
  FHeight := Value;
  Self.Recreate;
end;

procedure TCaret.SetPos(Value: TPoint);
begin
  if not FAllocated then Recreate;
  if FVisible then Hide;
  if SetCaretPos(Value.x, Value.y) then
    Self.FPos := Value;
  if FVisible then Show;
end;

procedure TCaret.SetVisible(Value: Boolean);
begin
  if not FAllocated then Recreate;
  if Value then Self.Show else Self.Hide;
  FVisible := Value;
end;

procedure TCaret.Show;
begin
  if not FAllocated then Recreate;
  if not FVisible then ShowCaret(Self.FParent.Handle);
  FVisible := True;
end;

{TCustomView}

constructor TCustomView.Create(Owner: TComponent);
begin
  inherited Create(Owner);

  FViewMode     := vmClear;
  GetMem(FBuffer, MaxStringLen);
  FUpperBuffer  := TStringList.Create;
  FLowerBuffer  := TStringList.Create;
  FBrush        := TBrush.Create;
  FActiveBrush  := TBrush.Create;
  FPageInfo     := TPageInfo.Create;
  FPageInfoMm   := TPageInfo.Create;
  FTempPageInfo := TPageInfo.Create;

  FBlockColor   := TColorSpec.Create(Self);
  FBlockColor.TextColor := clWhite;
  FBlockColor.BackColor := clNavy;

  FSelectedColor := TColorSpec.Create(Self);
  FSelectedColor.TextColor := clBlack;
  FSelectedColor.BackColor := clLime;
  FActiveBrush.Color := clLime;

  FSelectedBlockColor := TColorSpec.Create(Self);
  FSelectedBlockColor.TextColor := clWhite;
  FSelectedBlockColor.BackColor := clBlack;

  ParentColor        := False;
  Color              := clWindow;
  FBrush.Color       := Color;
  Font.Color         := clBlack;
  Font.Name          := 'Courier New';

  FLineIndex       := 0;
  FActiveLineShift := 0;
  FActivePosShift  := 0;
  FHorzShift       := 0;
  FLinesPerPage    := 60;
  FDOSPrintDevice  := 'LPT1';
  FPrintFontSize   := -1;
  FActiveDC := 0;
  FStartPos := 0;
  FMidPos   := 0;
  FEndPos   := 0;
  FFileSize := 0;
  FViewOptions     := [voTabs, voRowSelect, voCaret];
  FPrintingType    := ptEntireRange;

  FMarking    := False;
  FMarkType   := mtNone;
  FSilentMove := False;

  FMouseCaptured := False;

  with FTempPageInfo do
  begin
    FLeftMargin   := 20;
    FRightMargin  := 20;
    FTopMargin    := 50;
    FBottomMargin := 50;
    FHeight       := 250;
    FWidth        := 210;
    FPageSpacing  := 5;
    FLineGap      := 0;
  end;

  Font.Size := 10;
  Height    := 50;
  Width     := 50;
end;

destructor TCustomView.Destroy;
begin
  CloseFile;
  if FCaret <> nil then FCaret.Free;
  if FBrush <> nil then FBrush.Free;
  if FActiveBrush <> nil then FActiveBrush.Free;
  if FBuffer <> nil then FreeMem(FBuffer);
  if FUpperBuffer <> nil then FUpperBuffer.Free;
  if FLowerBuffer <> nil then FLowerBuffer.Free;
  if FPageInfo <> nil then FPageInfo.Free;
  if FPageInfoMm <> nil then FPageInfoMm.Free;
  if FTempPageInfo <> nil then FTempPageInfo.Free;
  if FBlockColor <> nil then FBlockColor.Free;
  if FSelectedColor <> nil then FSelectedColor.Free;
  if FSelectedBlockColor <> nil then FSelectedBlockColor.Free;
  FreeActiveDC;
  FreeClipboardHandle;
  inherited Destroy;
end;

procedure TCustomView.OpenFile(AFileName: String);
begin
  try
    CloseFile;
    FFileName := AFileName;

    if FFileName <> EmptyStr then
    begin
      FStream := TFileStream.Create(FFileName, fmOpenRead or
                                               fmShareDenyNone or
                                               fmShareCompat);
      FLineIndex   := 0;
      FGlobalLine  := 0;
      FActiveLineShift := 0;
      FActivePosShift  := 0;
      FHorzShift       := 0;
      FActiveDC        := 0;
      InitialFillBuffers;
      CancelMarking;
      UpdateCaret;
    end;
    Invalidate;
    Update;
  except
    CloseFile;
    raise;
  end;
end;

procedure TCustomView.CloseFile;
begin
  FUpperBuffer.Clear;
  FLowerBuffer.Clear;
  FLineIndex := 0;
  FActiveLineShift := 0;
  FStartPos := 0;
  FMidPos := 0;
  FEndPos := 0;
  FGlobalLine := 0;
  if FStream <> nil then FStream.Free;
  FStream := nil;
  FFileName := '';
  if FCaret <> nil then FCaret.Remove;
  Invalidate;
end;

procedure TCustomView.SaveAs(AFileName: String);
var
  SavePos: Integer;
  SaveCursor: Integer;
begin
  if FStream = nil then exit;

  SavePos := FStream.Position;
  FStream.Position := 0;

  SaveCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  try
    with TFileStream.Create(AFileName, fmCreate) do
    begin
      CopyFrom(FStream, FStream.Size);
      Free;
    end;
  finally
    FStream.Position := SavePos;
    Screen.Cursor := SaveCursor;
  end;
end;

procedure TCustomView.CMFontChanged(var Msg: TMessage);
var
  Shift: Integer;
begin
  inherited;
  if (not (csLoading in ComponentState)) and (Parent <> nil) then
  begin
    RecalcLineSize;
    if (FActiveLineShift+1) * FLineHeight >= ClientHeight then
    begin
      Shift := ((FActiveLineShift+1) * FLineHeight - ClientHeight) div FLineHeight + 1;
      Dec(FActiveLineShift, Shift);
      LineIndex := LineIndex + Shift;
    end;
  end;
  UpdateCaret;
end;

procedure TCustomView.RecalcLineSize;
var
  TM: TTextMetric;
  DC: HDC;
begin
  DC := ActiveDC;
  SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, TM);
  FLineHeight := TM.tmHeight;
  if FCaret <> nil then
    FCaret.Height := FLineHeight;
end;

function TCustomView.ActiveDC: HDC;
begin
  if Parent <>  nil then
   if FActiveDC = 0 then
      FActiveDC := GetDC(Handle);
  Result := FActiveDC;
end;

procedure TCustomView.FreeActiveDC;
begin
  if Parent = nil then exit;

  if FActiveDC <> 0 then Windows.ReleaseDC(Handle, FActiveDC);
  FActiveDC := 0;
end;

procedure TCustomView.WMPaint(var Msg: TWMPaint);
var
  PS: TPaintStruct;
begin
  BeginPaint(Handle, PS);
  try
    with PS do
    begin
      InvalidateView(rcPaint);
      RedrawView(HDC);
      UpdateActiveLine(False);
    end;
  finally
    EndPaint(Handle, PS);
  end;
end;

procedure TCustomView.KeyDown(var Key: Word; ShiftState: TShiftState);
begin
  inherited;

  if csDesigning in ComponentState then exit;

  case Key of
    VK_DOWN: MoveLineDown;
    VK_UP:   MoveLineUp;
    VK_RIGHT:
      begin
        if ssCtrl in ShiftState then
          MoveWordRight
        else
          MoveCaretRight;
      end;
    VK_LEFT:
      begin
        if ssCtrl in ShiftState then
          MoveWordLeft
        else
          MoveCaretLeft;
      end;
    VK_PRIOR: MovePageUp;
    VK_NEXT:  MovePageDown;
    VK_HOME:
      if ssCtrl in ShiftState then
        MoveTop
      else
        MoveLineHome;
    VK_END:
      if ssCtrl in ShiftState then
        MoveBottom
      else
        MoveLineEnd;
    VK_INSERT:
      if ssCtrl in ShiftState then
        CopyToClipboard;
  end;
end;

procedure TCustomView.KeyUp(var Key: Word; ShiftState: TShiftState);
begin
  inherited;
  if csDesigning in ComponentState then exit;
  case Key of
    VK_SHIFT:   FinishMarking(mtHorizontal);
    VK_MENU: FinishMarking(mtVertical);
  end;
end;

procedure TCustomView.WMGetDlgCode(var Msg: TWMGetDlgCode);
begin
  if csDesigning in ComponentState then exit;
  Msg.Result := DLGC_WANTARROWS or DLGC_WANTCHARS;

  if voTabs in FViewOptions then
    Msg.Result := Msg.Result or DLGC_WANTTAB;

end;

procedure TCustomView.InvalidateView(Rect: TRect);
begin
  FRect := Rect;
  EnumVisibleLines(InvalidateLine);
end;

procedure TCustomView.Counter(S: TStrings;
                              LocalIndex: Integer;
                              GlobalIndex: Integer;
                              R: TRect);
begin
  Inc(FVisibleLinesCount);
end;

procedure TCustomView.InvalidateLine(S: TStrings;
                                     LocalIndex: Integer;
                                     GlobalIndex: Integer;
                                     R: TRect);
begin
  if ((R.Top >= FRect.Top) or ((R.Top < FRect.Top) and (R.Bottom >= FRect.Top)))
     and (R.Top <= FRect.Bottom) then
    S.Objects[LocalIndex] := TObject(1);
end;

function TCustomView.EnumVisibleLines(LineProcessor: TLineProcessor): Integer;
var
  R: TRect;
  i: Integer;
  l: Integer;
begin

  Result := 0;

  if FStream = nil then exit;

  i := FLineIndex;
  R := Rect(0, 0, ClientWidth, FLineHeight);
  l := FUpperBuffer.Count;

  while (R.Top < ClientHeight) and (R.Bottom <= ClientHeight) do
  begin
    if i < l then
      LineProcessor(FUpperBuffer, i, FGlobalLine + i - FLineIndex, R)
    else
      if (i - l < FLowerBuffer.Count) then
        LineProcessor(FLowerBuffer, i - l, FGlobalLine + i - FLineIndex, R)
      else
        exit;

    Inc(i);
    Inc(R.Top, FLineHeight);
    Inc(R.Bottom, FLineHeight);
    Inc(Result);
  end;
end;

procedure TCustomView.DrawLine(S: TStrings;
                               LocalIndex: Integer;
                               GlobalIndex: Integer;
                               R: TRect);
var
  R1: TRect;
  sz: TSize;
  SaveBackColor: TColor;
  SaveTextColor: TColor;
  i1, i2: Integer;
begin

  if (S.Objects[LocalIndex] = TObject(0)) or
     ((S.Objects[LocalIndex] = TObject(1)) and
       (LocalIndex = FActiveLineShift + FLineIndex) and (voRowSelect in FViewOptions)) then exit;

  Inc(R.Left, FHorzShift);

  R1 := R;

  if (FMarkType = mtNone) or
     (
      (FMarkType <> mtNone) and
      (
       (GlobalIndex < FMarkBegin.Line) or
       (GlobalIndex > FMarkEnd.Line)
      )
     ) then
  begin
    GetTextExtentPoint32(FDC,
                         @(S.Strings[LocalIndex][1]),
                         Length(S.Strings[LocalIndex]),
                         sz);
    DrawText(FDC, @(S.Strings[LocalIndex][1]), Length(S.Strings[LocalIndex]), R,
             DT_LEFT or
             DT_BOTTOM or
             DT_NOPREFIX or
             DT_SINGLELINE or
             DT_EXTERNALLEADING
            );

    if FViewMode = vmClear then
    begin
      Inc(R1.Left, sz.cx);
      if R1.Left < R1.Right then
        FillRect(FDC, R1, FBrush.Handle);
    end;
  end
  else
  begin
    if (GlobalIndex = FMarkBegin.Line) or
       (
        (GlobalIndex > FMarkBegin.Line) and
        (GlobalIndex <= FMarkEnd.Line) and
        (FMarkType = mtVertical)
       ) then
      i1 := Min(FMarkBegin.Col, Length(S.Strings[LocalIndex]))
    else
      i1 := 0;

    if (GlobalIndex = FMarkEnd.Line) or
       (
        (GlobalIndex >= FMarkBegin.Line) and
        (GlobalIndex < FMarkEnd.Line) and
        (FMarkType = mtVertical)
       ) then
      i2 := Min(Length(S.Strings[LocalIndex]), FMarkEnd.Col)
    else
      i2 := Length(S.Strings[LocalIndex]);

    if i1 > 0 then
    begin
      GetTextExtentPoint32(FDC,
                           @(S.Strings[LocalIndex][1]),
                           i1,
                           sz);

      DrawText(FDC,
               @(S.Strings[LocalIndex][1]),
               i1,
               R,
               DT_LEFT or
               DT_BOTTOM or
               DT_NOCLIP or
               DT_NOPREFIX or
               DT_SINGLELINE or
               DT_EXTERNALLEADING
              );
      Inc(R1.Left, sz.cx);
      Inc(R.Left, sz.cx);
    end;


    if Length(S.Strings[LocalIndex]) > 0 then
    begin
      GetTextExtentPoint32(FDC,
                           @(S.Strings[LocalIndex][i1+1]),
                           i2 - i1,
                           sz);

      if FBrush.Color = SelectedColor.BackColor then
      begin
        SaveBackColor := SetBkColor(FDC, ColorToRGB(FSelectedBlockColor.BackColor));
        SaveTextColor := SetTextColor(FDC, ColorToRGB(FSelectedBlockColor.TextColor));
      end
      else
      begin
        SaveBackColor := SetBkColor(FDC, ColorToRGB(FBlockColor.BackColor));
        SaveTextColor := SetTextColor(FDC, ColorToRGB(FBlockColor.TextColor));
      end;

      DrawText(FDC,
               @(S.Strings[LocalIndex][i1+1]),
               i2-i1,
               R,
               DT_LEFT or
               DT_BOTTOM or
               DT_NOCLIP or
               DT_NOPREFIX or
               DT_SINGLELINE or
               DT_EXTERNALLEADING
              );
      Inc(R1.Left, sz.cx);
      Inc(R.Left, sz.cx);

      SetBkColor(FDC, SaveBackColor);
      SetTextColor(FDC, SaveTextColor);
    end;

    if (i2 < Length(S.Strings[LocalIndex])) and (i2 >= 0) then
    begin
      GetTextExtentPoint32(FDC,
                           @(S.Strings[LocalIndex][i2+1]),
                           Length(S.Strings[LocalIndex])-i2,
                           sz);

      DrawText(FDC,
               @(S.Strings[LocalIndex][i2+1]),
               Length(S.Strings[LocalIndex])-i2,
               R,
               DT_LEFT or
               DT_BOTTOM or
               DT_NOCLIP or
               DT_NOPREFIX or
               DT_SINGLELINE or
               DT_EXTERNALLEADING
              );
      Inc(R1.Left, sz.cx);
    end;

    if R1.Left < R1.Right then
      FillRect(FDC, R1, FBrush.Handle);
  end;

  S.Objects[LocalIndex] := TObject(0);
end;

procedure TCustomView.RedrawView(DC: HDC);
var
  OldFont: THandle;
  OldBrush: THandle;
  R: TRect;
  TotalLines: Integer;
begin
  if Parent = nil then exit;

  OldFont := SelectObject(DC, Font.Handle);
  OldBrush := SelectObject(DC, FBrush.Handle);
  SetTextColor(DC, ColorToRGB(Font.Color));
  SetBkColor(DC, ColorToRGB(FBrush.Color));
  FDC := DC;
  try
    TotalLines := EnumVisibleLines(DrawLine);
    R := Rect(0, TotalLines * FLineHeight, ClientWidth, ClientHeight);
    FillRect(DC, R, FBrush.Handle);
    UpdateActiveLine(False);
  finally
    SelectObject(DC, OldFont);
    SelectObject(DC, OldBrush);
  end;
end;

function TCustomView.ReadLineBack(var S: String;
                                  ReadingState: TLineReadingState): Boolean;
var
  i: Integer;
  p: PChar;
  l: Integer;
begin
  i := FStream.Position;
  l := Min(MaxStringLen-1, i);
  FStream.Position := i - l;
  l := FStream.Read(FBuffer[0], l);
  if FBuffer[l-1] = #10 then
    FBuffer[l-2] := #0
  else
    FBuffer[l] := #0;

  p := StrRScan(FBuffer, #10);

  if (p = nil) and (l > 0) then
    p := FBuffer - 1;

  if p <> nil then
  begin
    p := p + 1;
    S := StrPas(p);
    FStream.Position := i - (l - (p - FBuffer));
  end;

  if Assigned(FStringEncoder) then
    FStringEncoder(Self, S, ReadingState);

  Result := p <> nil;
end;

function TCustomView.ReadLineForward(var S: String;
                                     ReadingState: TLineReadingState): Boolean;
var
  p: PChar;
  i: Integer;
  l: Integer;
begin
  i := FStream.Position;
  l := FStream.Read(FBuffer[0], MaxStringLen-1);
  FBuffer[l] := #0;
  p := StrPos(FBuffer, PChar(CRLF));

  if (p = nil) and (l > 0) then
    p := FBuffer + l;

  if p <> nil then
  begin
    p[0] := #0;
    S := StrPas(FBuffer);
    FStream.Position := i + Min(p - FBuffer + 2, l);
  end;
  
  if Assigned(FStringEncoder) then
    FStringEncoder(Self, S, ReadingState);

  Result := p <> nil;
end;

procedure TCustomView.SwapBuffers;
var
  S: TStrings;
begin
  S := TStringList.Create;
  try
    S.Assign(FUpperBuffer);
    FUpperBuffer.Assign(FLowerBuffer);
    FLowerBuffer.Assign(S);
    FWasSwap := True;
  finally
    S.Free;
  end;
end;

procedure TCustomView.InitialFillBuffers;
var
  i: Integer;
  S: String;
  SaveCursor: Integer;
begin

  SaveCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  try
    FStartPos := 0;
    FStream.Position := FStartPos;

    for i := 0 to InitialBufLength-1 do
      if ReadLineForward(S, lrsInitializing) then
        FUpperBuffer.AddObject(S, TObject(0))
      else
        break;

    FMidPos := FStream.Position;

    for i := 0 to InitialBufLength-1 do
      if ReadLineForward(S, lrsInitializing) then
        FLowerBuffer.AddObject(S, TObject(0))
      else
        break;

    FEndPos := FStream.Position;
  finally
    Screen.Cursor := SaveCursor;
  end;
end;

function TCustomView.UpdateText(NeededShift: Integer): Integer;
var
  i: Integer;
  S: String;
  SaveCursor: Integer;
begin

  FWasSwap := False;
  SaveCursor := Screen.Cursor;
  Result := NeededShift;

  if (FEndPos < FStream.Size) and
     (FUpperBuffer.Count + FLowerBuffer.Count < 2*InitialBufLength) then
  begin
    Screen.Cursor := crHourGlass;
    try
      FStream.Position := FStartPos;

      FUpperBuffer.Clear;

      for i := 0 to InitialBufLength - 1 do
        if ReadLineForward(S, lrsWindowSwap) then
          FUpperBuffer.AddObject(S, TObject(0))
        else
          break;

      FMidPos := FStream.Position;
      FEndPos := FMidPos;


      FLowerBuffer.Clear;
      for i := 0 to InitialBufLength - 1 do
        if ReadLineForward(S, lrsWindowSwap) then
          FLowerBuffer.AddObject(S, TObject(0))
        else
          break;

      FEndPos := FStream.Position;
    finally
      Screen.Cursor := SaveCursor;
    end;

    Result := UpdateText(NeededShift);
    exit;
  end
  else if (FLineIndex >= InitialBufLength) and (FEndPos < FStream.Size) then
  begin
    Screen.Cursor := crHourGlass;
    try
      SwapBuffers;
      FStartPos := FMidPos;
      FMidPos := FEndPos;

      FStream.Position := FEndPos;
      FLowerBuffer.Clear;
      for i := 0 to InitialBufLength-1 do
        if ReadLineForward(S, lrsWindowSwap) then
          FLowerBuffer.AddObject(S, TObject(0))
        else
          break;

      FEndPos := FStream.Position;

      FLineIndex := FLineIndex - InitialBufLength;
    finally
      Screen.Cursor := SaveCursor;
    end;
  end
  else if FLineIndex < 0 then
  begin
    if FStartPos = 0 then
    begin
      Result := -FLineIndex+NeededShift;
      FLineIndex := 0;
      exit;
    end;

    Screen.Cursor := crHourGlass;
    try
      SwapBuffers;
      FEndPos := FMidPos;
      FMidPos := FStartPos;

      FStream.Position := FMidPos;

      FUpperBuffer.Clear;

      for i := 0 to InitialBufLength-1 do
        if ReadLineBack(S, lrsWindowSwap) then
          FUpperBuffer.InsertObject(0, S, TObject(0))
        else
          break;
      FStartPos := FStream.Position;

      Result := Max(-FLineIndex+NeededShift-FUpperBuffer.Count, NeededShift);
      FLineIndex := Max(0, Min(FUpperBuffer.Count, FLineIndex + FUpperBuffer.Count));

    finally
      Screen.Cursor := SaveCursor;
    end;
  end;

  if (FEndPos >= FStream.Size) and
     ((FUpperBuffer.Count + FLowerBuffer.Count - FLineIndex) < FVisibleLinesCount) then
  begin
    Result := FUpperBuffer.Count + FLowerBuffer.Count - FVisibleLinesCount - FLineIndex + NeededShift;
    FLineIndex := FUpperBuffer.Count + FLowerBuffer.Count - FVisibleLinesCount;
    exit;
  end;
end;

procedure TCustomView.ScrollArea(Dx, Dy: Integer);
var
  R: TRect;
  IR, IR1: TRect;
  TextHeight: Integer;
begin

  if FSilentMove then exit;

  TextHeight := ClientHeight - (ClientHeight mod FLineHeight);

  if Abs(Dy) > FLineHeight then
  begin
    UpdateActiveLine(True);
    R := Rect(0,
              FActiveLineShift * FLineHeight,
              ClientWidth,
              (FActiveLineShift + 1) * FLineHeight);
    InvalidateRect(Handle, @R, False);
    RedrawView(ActiveDC);
  end;

  R := ClientRect;
  if Dy < 0 then
  begin
    IR := Rect(0, TextHeight + Dy, ClientWidth, ClientHeight);
    IR1 := Rect(0, 0, 0, 0);
    Dec(R.Top, Dy);
    if Abs(Dy) <= FLineHeight then
      R.Bottom := TextHeight + Dy;
  end
  else
  begin
    IR := Rect(0, 0, ClientWidth, Dy);
    IR1 := Rect(0, TextHeight, ClientWidth, ClientHeight);
    Inc(R.Top, Dy);
    R.Bottom := TextHeight;
  end;

  if Dy > 0 then
    Dec(R.Bottom, ClientHeight mod FLineHeight);

  ScrollWindowEx(Handle,
                 Dx,
                 Dy,
                 @R,
                 nil,
                 0,
                 nil,
                 SW_INVALIDATE);

  if Abs(Dy) > FLineHeight then UpdateActiveLine(True);

  InvalidateRect(Handle, @IR, False);
  InvalidateRect(Handle, @IR1, False);
  Update;
end;

procedure TCustomView.SetLineIndex(Value: Integer);
var
  Shift: Integer;
begin
  GetVisibleLinesCount;
  Shift := Value - FLineIndex;
  FLineIndex := Value;
  Shift := UpdateText(Shift);

  if Shift <> 0 then
  begin
    Inc(FGlobalLine, Shift);
    UpdateMarking;
    ScrollArea(0, -Shift * FLineHeight);
  end
end;

function TCustomView.GetVisibleLinesCount: Integer;
begin
  FVisibleLinesCount := 0;
  EnumVisibleLines(Counter);
  Result := FVisibleLinesCount;
end;

procedure TCustomView.TextLineUp;
begin
  if FStream = nil then exit;
  LineIndex := LineIndex - 1;
end;

procedure TCustomView.TextLineDown;
begin
  if FStream = nil then exit;
  LineIndex := LineIndex + 1;
end;

procedure TCustomView.InternalMovePageUp;
var
  Old: Integer;
begin
  if FStream = nil then exit;

  CheckMarking;
  if not (voCaret in ViewOptions) then
    FActivePosShift := 0;

  Old := LineIndex;
  LineIndex := LineIndex - Self.ClientHeight div FLineHeight;

  if Old = LineIndex then
    if FMarking then
    begin
      FActiveLineShift := 0;
      UpdateMarking;
      Invalidate;
      Update;
    end
    else
    begin
      UpdateActiveLine(True);
      FActiveLineShift := 0;
      UpdateActiveLine(False);
    end

end;

procedure TCustomView.InternalMovePageDown;
var
  Old: Integer;
begin
  if FStream = nil then exit;

  CheckMarking;
  if not (voCaret in ViewOptions) then
    FActivePosShift := 0;

  Old := LineIndex;
  LineIndex := LineIndex + Self.ClientHeight div FLineHeight;

  if Old = LineIndex then
    if FMarking then
    begin
      FActiveLineShift := FVisibleLinesCount - 1;
      UpdateMarking;
      Invalidate;
      Update;
    end
    else
    begin
      UpdateActiveLine(True);
      FActiveLineShift := FVisibleLinesCount - 1;
      UpdateActiveLine(False);
    end;
end;

procedure TCustomView.MovePageUp;
begin
  Perform(WM_VSCROLL, SB_PAGEUP, 0);
end;

procedure TCustomView.MovePageDown;
begin
  Perform(WM_VSCROLL, SB_PAGEDOWN, 0);
end;

procedure TCustomView.MoveTop;
begin
  Perform(WM_VSCROLL, SB_TOP, 0);
end;

procedure TCustomView.MoveBottom;
begin
  Perform(WM_VSCROLL, SB_BOTTOM, 0);
end;

procedure TCustomView.MoveLineUp;
begin
  Perform(WM_VSCROLL, SB_LINEUP, 0);
end;

procedure TCustomView.MoveLineDown;
begin
  Perform(WM_VSCROLL, SB_LINEDOWN, 0);
end;

procedure TCustomView.MoveLineHome;
begin
  Perform(WM_HSCROLL, SB_TOP, 0);
end;

procedure TCustomView.MoveLineEnd;
begin
  Perform(WM_HSCROLL, SB_BOTTOM, 0);
end;

procedure TCustomView.MoveCaretLeft;
begin
  Perform(WM_HSCROLL, SB_LINELEFT, 0);
end;

procedure TCustomView.MoveCaretRight;
begin
  Perform(WM_HSCROLL, SB_LINERIGHT, 0);
end;

procedure TCustomView.MoveWordLeft;
begin
  Perform(WM_HSCROLL, SB_PAGELEFT, 0);
end;

procedure TCustomView.MoveWordRight;
begin
  Perform(WM_HSCROLL, SB_PAGERIGHT, 0);
end;


procedure TCustomView.SetPageInfoMm(Value: TPageInfo);
begin
  FPageInfoMm := Value;
  RecalcMargins;
end;

procedure TCustomView.RecalcMargins;
var
  LogX: Integer;
  LogY: Integer;
begin
  LogX := GetDeviceCaps(ActiveDC, LOGPIXELSX);
  LogY := GetDeviceCaps(ActiveDC, LOGPIXELSY);

  with FPageInfoMm do
  begin
    FPageInfo.FLeftMargin   := (LogX * FLeftMargin * 10) div 254;
    FPageInfo.FRightMargin  := (LogX * FRightMargin * 10) div 254;
    FPageInfo.FWidth        := (LogX * FWidth * 10) div 254;
    //
    FPageInfo.FTopMargin    := (LogY * FTopMargin * 10) div 254;
    FPageInfo.FBottomMargin := (LogY * FBottomMargin * 10) div 254;
    FPageInfo.FHeight       := (LogY * FHeight * 10) div 254;
    FPageInfo.FPageSpacing  := (LogY * FPageSpacing * 10) div 254;
    FPageInfo.FLineGap  := (LogY * FLineGap * 10) div 254;
  end;
end;

procedure TCustomView.CMShowingChanged(var Msg: TMessage);
begin
  inherited;
  if Visible then
    FPageInfoMm.Assign(FTempPageInfo);
  FActiveDC := 0;
  UpdateCaret;
end;

procedure TCustomView.UpdateActiveLine(HideIt: Boolean);
var
  R: TRect;
  P: TPoint;
  l: Integer;
  S: TStrings;
  TempBrush: TBrush;
  OldFont: THandle;
begin

  if (FStream = nil) or FSilentMove or (FUpperBuffer.Count = 0) then exit;

  if not (voRowSelect in FViewOptions) then HideIt := True;

  if FCaret.Visible then FCaret.Hide;
  l := FLineIndex + FActiveLineShift;

  TempBrush := nil;

  R := Rect(0,
            FActiveLineShift * FLineHeight,
            ClientWidth,
            (FActiveLineShift + 1) * FLineHeight);

  if l >= FUpperBuffer.Count then
  begin
    l := l - FUpperBuffer.Count;
    S := FLowerBuffer
  end
  else
    S := FUpperBuffer;

  FDC := ActiveDC;

  if not HideIt then
  begin
    TempBrush := FBrush;
    FBrush := FActiveBrush;
    SetTextColor(FDC, SelectedColor.TextColor);
  end
  else
    SetTextColor(FDC, Font.Color);

  SetBkColor(FDC, ColorToRGB(FBrush.Color));
  S.Objects[l] := TObject(-1);

  P := R.TopLeft;
  Inc(P.x, CalcPosShift(S[l], FActivePosShift) + FHorzShift);
  FCaret.Pos := P;

  OldFont := SelectObject(FDC, Font.Handle);
  try
    DrawLine(S, l, FGlobalLine + FActiveLineShift, R);
  finally
    if not HideIt then
      FBrush := TempBrush;
    SelectObject(FDC, OldFont);
  end;

  if (csDesigning in ComponentState) or (not Focused) or
     (not (voCaret in FViewOptions))then
    FCaret.Remove
  else
    FCaret.Visible := True;
end;

procedure TCustomView.InternalMoveLineDown;
begin
  CheckMarking;
  if not (voCaret in ViewOptions) then
    FActivePosShift := 0;

  if (FActiveLineShift < GetVisibleLinesCount-1) and
     ((voRowSelect in FViewOptions) or (voCaret in FViewOptions)) then
    if FMarking then
    begin
      UpdateActiveLine(True);
      Inc(FActiveLineShift);
      UpdateMarking;
      Dec(FActiveLineShift);
      UpdateActiveLine(True);
      Inc(FActiveLineShift);
      UpdateActiveLine(False);
    end
    else
    begin
      UpdateActiveLine(True);
      Inc(FActiveLineShift);
      UpdateActiveLine(False);
    end
  else
    TextLineDown;
end;

procedure TCustomView.InternalMoveLineUp;
begin
  CheckMarking;

  if not (voCaret in ViewOptions) then
    FActivePosShift := 0;

  if (FActiveLineShift > 0) and
     ((voRowSelect in FViewOptions) or (voCaret in FViewOptions)) then
    if FMarking then
    begin
      UpdateActiveLine(True);
      Dec(FActiveLineShift);
      UpdateMarking;
      Inc(FActiveLineShift);
      UpdateActiveLine(True);
      Dec(FActiveLineShift);
      UpdateActiveLine(False);
    end
    else
    begin
      UpdateActiveLine(True);
      Dec(FActiveLineShift);
      UpdateActiveLine(False);
    end
  else
    TextLineUp;
end;

procedure TCustomView.ScrollText(Shift: Integer);
var
  RealShift: Integer;
begin
  RealShift := Min(Shift, -FHorzShift);
  Inc(FHorzShift, RealShift);
  Self.ScrollArea(RealShift, 0);
end;

procedure TCustomView.UpdateCaret;
begin
  if Parent = nil then exit;

  if FCaret = nil then FCaret := TCaret.Create(Self, FLineHeight);
  RecalcLineSize;

  if (csDesigning in ComponentState) or (not Focused) or
     (not (voCaret in FViewOptions))then
    FCaret.Remove
  else
    FCaret.Visible := True;
end;

procedure TCustomView.WMLButtonDown(var Msg: TWMLButtonDown);
begin
  inherited;
  SetForegroundWindow(Handle);
  SetFocus;
  if csDesigning in ComponentState then exit;
  SetCapture(Handle);
  FMousePoint := Point(Msg.XPos, Msg.YPos);
  FTimerID := SetTimer(Handle, 0, 25, nil);
  SetLineByPoint(FMousePoint);
  FMouseCaptured := True;
end;

procedure TCustomView.SetLineByPoint(Pt: TPoint);
var
  NewShift: Integer;
  Old: Integer;
begin
  if FStream = nil then exit;

  FPoint := Pt;
  if Pt.y < 0 then
    FNewLine := -1;

  FNewLine := -1;
  EnumVisibleLines(FindLine);

  if FNewLine = -1 then
    if (Pt.y < 0) and (FActiveLineShift <> 0) then
      FNewLine := FGlobalLine//FLineIndex
    else if (Pt.y > ClientHeight) and (FActiveLineShift < FVisibleLinesCount-1) then
      FNewLine := FGlobalLine {FLineIndex} + FVisibleLinesCount-1;

  if FNewLine <> -1 then
  begin
    if FNewLine < FGlobalLine {FLineIndex} then
      NewShift := FUpperBuffer.Count - FGlobalLine{FLineIndex} + FNewLine
    else
      NewShift := FNewLine - FGlobalLine{FLineIndex};

    FActivePosShift := Max(CalcRealShift(NewShift, Pt.x), 0);

    UpdateActiveLine(True);
    FActiveLineShift := NewShift;
    UpdateActiveLine(False);
    CheckMarking;

    if FMarking then
    begin
      UpdateMarking;
      if not Equel(FMarkBegin, FMarkEnd) then
      begin
        Invalidate;
        Update;
      end;
    end;
  end
  else if Pt.y < 0 then
  begin
    CheckMarking;
    if not (voCaret in ViewOptions) then
      FActivePosShift := 0;

    Old := LineIndex;
    LineIndex := LineIndex - Min(Self.ClientHeight div FLineHeight, (-Pt.y div FLineHeight) + 1);

    if Old = LineIndex then
      if FMarking then
      begin
        FActiveLineShift := 0;
        UpdateMarking;
        Invalidate;
        Update;
      end
      else
      begin
        UpdateActiveLine(True);
        FActiveLineShift := 0;
        UpdateActiveLine(False);
      end;
  end
  else if Pt.y > ClientHeight then
  begin
    CheckMarking;
    if not (voCaret in ViewOptions) then
      FActivePosShift := 0;

    Old := LineIndex;

    LineIndex := LineIndex + Min(Self.ClientHeight div FLineHeight,
                                 ((Pt.y - ClientHeight) div FLineHeight) + 1);
    if Old = LineIndex then
      if FMarking then
      begin
        FActiveLineShift := FVisibleLinesCount - 1;
        UpdateMarking;
        Invalidate;
        Update
      end
      else
      begin
        UpdateActiveLine(True);
        FActiveLineShift := FVisibleLinesCount - 1;
        UpdateActiveLine(False);
      end;
  end;
end;

procedure TCustomView.FindLine(S: TStrings;
                               LocalIndex: Integer;
                               GlobalIndex: Integer;
                               R: TRect);
begin
  if FNewLine <> -1 then exit;
  if ((FPoint.y < R.Top) and (LocalIndex = 0) and (S = FUpperBuffer) and (FStartPos = 0)) or
     ((FPoint.y >= R.Top) and (FPoint.y <= R.Bottom)) or
     ((FPoint.y > R.Bottom) and ((LocalIndex = S.Count-1) and
                                 ((S = FLowerBuffer) or
                                  ((S = FUpperBuffer) and (FLowerBuffer.Count = 0))))) then
     FNewLine := GlobalIndex{LocalIndex};
end;

procedure TCustomView.WMLButtonUp(var Msg: TWMLButtonUp);
begin
  inherited;
  if csDesigning in ComponentState then exit;
  FinishMarking(mtAny);
  KillTimer(Handle, FTimerID);
  ReleaseCapture;
  FMouseCaptured := False;
end;

procedure TCustomView.WMMouseMove(var Msg: TWMMouseMove);
begin
  inherited;
  if csDesigning in ComponentState then exit;
  FMousePoint := Point(Msg.XPos, Msg.YPos);
  if FMouseCaptured then
    SetLineByPoint(FMousePoint);
end;

procedure TCustomView.WMTimer(var Msg: TWMTimer);
var
  R: TRect;
begin
  if csDesigning in ComponentState then exit;
  R := ClientRect;
  if FMouseCaptured and (not PtInRect(R, FMousePoint)) then
  begin
    SetLineByPoint(FMousePoint);
    if FMousePoint.x > ClientWidth then
      InternalMoveCaret(1, False)
    else if FMousePoint.x < 0 then
      InternalMoveCaret(-1, False);
  end;
end;

procedure TCustomView.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    WindowClass.Style := WindowClass.Style or CS_VREDRAW or CS_HREDRAW;
//    Style := Style or WS_VSCROLL or WS_HSCROLL;
  end;
end;

procedure TCustomView.InternalMoveTop;
var
  Dy: Integer;
begin
  if FStream = nil then exit;
  
  CheckMarking;
  if not (voCaret in ViewOptions) then
    FActivePosShift := 0;

  UpdateActiveLine(True);
  Dy := FGlobalLine * FLineHeight;
  FLineIndex := 0;
  FGlobalLine := 0;
  FActiveLineShift := 0;
  FUpperBuffer.Clear;
  FLowerBuffer.Clear;
  InitialFillBuffers;
  UpdateMarking;

  ScrollArea(0, Dy);

  if Dy = 0 then
  begin
    if FMarking then
    begin
      Invalidate;
      Update;
    end;
    UpdateActiveLine(False);
  end;

end;

procedure TCustomView.InternalMoveBottom;
var
  i: Integer;
  S: String;
  Old: Integer;
  SaveCursor: Integer;
begin
  if FStream = nil then exit;

  SaveCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;

  try
    CheckMarking;
    if not (voCaret in ViewOptions) then
      FActivePosShift := 0;

    CountLinesAtFile;
    GetVisibleLinesCount;

    UpdateActiveLine(True);
    FStream.Position := FStream.Size;

    FEndPos := FStream.Position;
    FMidPos := FEndPos;

    FUpperBuffer.Clear;
    FLowerBuffer.Clear;
    for i := 0 to FVisibleLinesCount-1 do
      if ReadLineBack(S, lrsWindowSwap) then
        FUpperBuffer.InsertObject(0, S, TObject(0))
      else
        break;

    FStartPos := FStream.Position;
    Old := FGlobalLine;
    FGlobalLine := FLinesCount - FUpperBuffer.Count;

    FActiveLineShift := FUpperBuffer.Count-1;
    FLineIndex := 0;
    UpdateMarking;
    ScrollArea(0, (Old - FGlobalLine)*FLineHeight);
    if (Old = FGlobalLine) and FMarking then
    begin
      Invalidate;
      Update;
    end;

    UpdateActiveLine(False);
  finally
    Screen.Cursor := SaveCursor;
  end;

end;

function CountChars(Buffer: PChar; Ch: Char): Integer; assembler;
asm
  push  edi
  push  ebx
  push  esi

  xor   esi, esi
  or    eax, eax
  je    @@1

  mov   ebx, eax
  xor   al, al
  mov   edi, ebx
  mov   ecx, 0FFFFFFFFh
  repne scasb
  not   ecx
  dec   ecx

  mov   al, dl
  mov   edi, ebx
@@2:
  repne scasb
  jne    @@1
  inc   esi
  jmp   @@2
@@1:
  mov   eax, esi
  pop   esi
  pop   ebx
  pop   edi
end;

procedure TCustomView.CountLinesAtFile;
var
  Buffer: PChar;
  l: Integer;
  SavePos: Integer;
  BufSize: Integer;
  HasLF: Boolean;
begin
  BufSize := 200*1024;
  l := 0;

  if FStream.Size <> FFileSize then
  begin
    HasLF := False;

    GetMem(Buffer, BufSize);
    SavePos := FStream.Position;
    FStream.Position := 0;
    FLinesCount := 0;

    try
      while True do
      begin
        if l <> 0 then  HasLF := Buffer[l-1] = #10;
        l := FStream.Read(Buffer[0], BufSize-1);
        if l = 0 then
        begin
          if HasLF then
            Dec(FLinesCount);
          break;
        end;
        Buffer[l] := #0;
        Inc(FLinesCount, CountChars(Buffer, #10));
      end;
      Inc(FLinesCount);
      FFileSize := FStream.Size;
    finally
      FreeMem(Buffer);
      FStream.Position := SavePos;
    end;
  end;
end;

var
  WordSeparator: array [0..8] of string = (',', '.', ' ', '=', '#', '!', '?', '>', '<');

  function IsWordBreak(Ch: Char): Boolean;
  var
    i: Integer;
  begin
    Result := False;
    for i := Low(WordSeparator) to High(WordSeparator) do
      if ch = WordSeparator[i] then
      begin
        Result := True;
        break;
      end;
  end;
function TCustomView.CalcLeftWordOffset(S: String; FromShift: Integer; WordNum: Integer): Integer;
var
  HasSeparator: Boolean;
  i: Integer;
  Sep: Boolean;
begin
  HasSeparator := not ((FromShift = 0) or ((FromShift > 0) and ( not IsWordBreak(S[FromShift+1]))));

  Result := 0;

  for i := FromShift to Length(S)-1 do
  begin
    Sep := IsWordBreak(S[i+1]);
    if (not Sep) and HasSeparator and (WordNum = 1) then
    begin
      Result := i;
      break;
    end
    else if Sep then
      HasSeparator := True
    else if (not Sep) and HasSeparator then
    begin
      HasSeparator := False;
      Dec(WordNum);
    end;
  end;

end;

function TCustomView.CalcRightWordOffset(S: String; FromShift: Integer; WordNum: Integer): Integer;
var
  HasWord: Boolean;
  i: Integer;
  Sep: Boolean;
begin
  HasWord := not ((FromShift = 0) or (FromShift = Length(S)) or ((FromShift < Length(S)) and (IsWordBreak(S[FromShift]))));

  Result := 0;

  if FromShift = 0 then exit;

  for i := FromShift-1 downto 0 do
  begin
    Sep := IsWordBreak(S[i+1]);
    if Sep and HasWord and (WordNum = 1) then
    begin
      Result := i+1;
      break;
    end
    else if not Sep then
      HasWord := True
    else if Sep and HasWord then
    begin
      HasWord := False;
      Dec(WordNum);
    end;
  end;

end;

function TCustomView.CalcPosShift(S: String; Shift: Integer): Integer;
var
  sz: TSize;
begin
  if Length(S) < Shift then S := Padr(S, Shift);
  if Length(S) = 0 then S := ReplicateChar(#20, Shift);
  GetTextExtentPoint32(ActiveDC, @(S[1]), Shift, sz);
  Result := sz.cx;
end;

procedure TCustomView.CMDesignHitTest(var Msg: TCMDesignHitTest);
begin
  Msg.Result := Longint(BOOL(False));
end;

procedure TCustomView.WMSetFocus(var Msg: TWMSetFocus);
begin
  inherited;
  if (FCaret <> nil) and ( not (csDesigning in ComponentState)) and
     (voCaret in FViewOptions) then
    UpdateActiveLine(False);
end;

procedure TCustomView.WMKillFocus(var Msg: TWMKillFocus);
begin
  inherited;

  if (FCaret <> nil) then
    FCaret.Remove;
end;

procedure TCustomView.WMDestroy(var Msg: TMessage);
begin
  KillTimer(Handle, FTimerID);
  inherited;
end;

procedure TCustomView.WMVScroll(var Msg: TWMVScroll);
begin
  inherited;
  case Msg.ScrollCode of
    SB_BOTTOM:   InternalMoveBottom;
    SB_TOP:      InternalMoveTop;
    SB_LINEUP:   InternalMoveLineUp;
    SB_LINEDOWN: InternalMoveLineDown;
    SB_PAGEUP:   InternalMovePageUp;
    SB_PAGEDOWN: InternalMovePageDown;
  end;
end;

procedure TCustomView.WMHScroll(var Msg: TWMHScroll);
begin
  inherited;
  case Msg.ScrollCode of
    SB_BOTTOM:    InternalMoveLineEnd;
    SB_TOP:       InternalMoveLineHome;
    SB_LINELEFT:  InternalMoveCaret(-1, True);
    SB_LINERIGHT: InternalMoveCaret(1, True);
    SB_PAGELEFT:  InternalMoveWord(-1);
    SB_PAGERIGHT: InternalMoveWord(1);
  end;
end;

procedure TCustomView.WMCopy(var Msg: TWMCopy);
begin
  Self.CopyToClipboard;
end;

procedure TCustomView.WMSize(var Msg: TWMSize);
var
  HeightToActiveLine: Integer;
  LinesToScroll: Integer;
begin
  HeightToActiveLine := (FActiveLineShift + 2) * FLineHeight;
  if Msg.Height <= FLineHeight then
    Msg.Height := FLineHeight
  else if Msg.Height < HeightToActiveLine then
  begin
    LinesToScroll := (HeightToActiveLine - Msg.Height) div FLineHeight;
    UpdateActiveLine(True);
    Dec(FActiveLineShift, LinesToScroll);
    LineIndex := LineIndex + LinesToScroll;
    UpdateActiveLine(False);
    inherited
  end
  else
    inherited;
end;

procedure TCustomView.SetViewOptions(Value: TViewOptions);
begin
  FViewOptions := Value;
  UpdateActiveLine(not (voRowSelect in FViewOptions));
end;

procedure TCustomView.CMColorChanged(var Msg: TMessage);
begin
  inherited;
  FBrush.Color := Color;
end;

procedure TCustomView.CMCtl3DChanged(var Msg: TMessage);
begin
  inherited;
end;

procedure TCustomView.WMEraseBckgnd(var Msg: TMessage);
begin
end;

procedure TCustomView.InternalMoveCaret(Count: Integer; MoveBehind: Boolean);
var
  l: Integer;
  RealShift: Integer;
  Shift: Integer;
  S: String;
  OldShift: Integer;
begin

  if (FStream = nil) or
     ((FUpperBuffer.Count = 0) and (FLowerBuffer.Count = 0)) then exit;

  CheckMarking;

  if not (voCaret in ViewOptions) then
  begin
    ScrollText(-Count * 20);
    exit;
  end;

  l := FLineIndex + FActiveLineShift;

  if l >= FUpperBuffer.Count then
    S := FLowerBuffer[l - FUpperBuffer.Count]
  else
    S := FUpperBuffer[l];

  OldShift := FActivePosShift;

  RealShift := Min(FActivePosShift, -Count);
  Dec(FActivePosShift, RealShift);

  if MoveBehind then
    PadR(S, FActivePosShift+1)
  else
    if FActivePosShift > Length(S) then
      FActivePosShift := OldShift;

  if FActivePosShift = OldShift then exit;

  UpdateMarking;

  if FMarking and (FMarkType = mtVertical) then
  begin
    Invalidate;
    Update;
  end;

  Shift := CalcPosShift(S, FActivePosShift);
  if Shift > ClientWidth - FHorzShift then
    ScrollText(ClientWidth - FHorzShift - Shift - 2)
  else if Shift + FHorzShift < 0 then
    ScrollText(-Shift - FHorzShift)
  else
    UpdateActiveLine(False);
end;

procedure TCustomView.InternalMoveWord(Count: Integer);
var
  Shift: Integer;
  l: Integer;
  S: String;
begin
  if FStream = nil then exit;

  CheckMarking;

  if not (voCaret in ViewOptions) then
  begin
    ScrollText(-Count * 20);
    exit;
  end;

  l := FLineIndex + FActiveLineShift;

  if l >= FUpperBuffer.Count then
    S := FLowerBuffer[l - FUpperBuffer.Count]
  else
    S := FUpperBuffer[l];

  if Count > 0 then
  begin
    Shift := CalcLeftWordOffset(S, FActivePosShift, Count);
    if Shift > FActivePosShift then
      InternalMoveCaret(Shift - FActivePosShift, False)
    else if FActivePosShift < Length(S) then
      InternalMoveCaret(Length(S) - FActivePosShift, False);
  end
  else if Count < 0 then
  begin
    Shift := CalcRightWordOffset(S, FActivePosShift, -Count);
    if Shift < FActivePosShift then
      InternalMoveCaret(Shift - FActivePosShift, False)
    else if FActivePosShift > 0 then
      InternalMoveCaret( -FActivePosShift, False);
  end;
end;

function RPos(SubStr, Str: String): Integer;
var
  l, i: Integer;
begin
  Result := 0;
  i := 0;
  while True do
  begin
    l := Pos(SubStr, Str);
    if l = 0 then break;

    Inc(i, l);

    Result := i;                                // aabbbbaaccaa -> bbbbaaccaa -> ccaa

    Inc(i, Length(SubStr)-1);

    Str := Copy(Str, l + Length(SubStr), Length(Str));
  end;

end;

function TCustomView.FindText(Text: String; FindOptions: TViewFindOptions): Boolean;
var
{
  SaveStartPos: Integer;
  SaveMidPos: Integer;
  SaveEndPos: Integer;
  SaveLineIndex: Integer;
  SaveShift: Integer;
  SavePos: Integer;
}
  StartShift: Integer;
  WhereFoundSomething: Integer;
  Old: Integer;
  S: String;
  SaveCursor: Integer;

  function SearchScopeForward: Integer;
  var
    l: Integer;
    uc: Integer;
    i: Integer;
    P: PChar;
    PRes: PChar;
  begin
    uc := FUpperBuffer.Count;
    Result := 0;

    while True do
    begin
      l := FLineIndex + StartShift;

      if FGlobalLine + StartShift = FLastFoundLine then
        i := FLastFoundCol
      else
        i := 0;

      if l >= uc then
        P := PChar(FLowerBuffer[l-uc])
      else
        P := PChar(FUpperBuffer[l]);

      PRes := StrPos(P+i, PChar(Text));
      if PRes = nil then
        Result := 0
      else
        Result := PRes - P + 1;

      if (StartShift = FVisibleLinesCount-1) or ((Result <> 0)) then break;

      Inc(StartShift);
    end;
  end;

  function SearchScopeForwardU: Integer;
  var
    l: Integer;
    uc: Integer;
    i: Integer;
    S: String;
    P: PChar;
    PRes: PChar;
  begin
    uc := FUpperBuffer.Count;
    Result := 0;

    while True do
    begin
      l := FLineIndex + StartShift;

      if FGlobalLine + StartShift = FLastFoundLine then
        i := FLastFoundCol
      else
        i := 0;

      if l >= uc then
        S := AnsiUpperCase(FLowerBuffer[l-uc])
      else
        S := AnsiUpperCase(FUpperBuffer[l]);

      Text := AnsiUpperCase(Text);

      P := PChar(S);

      PRes := StrPos(P+i, PChar(Text));
      if PRes = nil then
        Result := 0
      else
        Result := PRes - P + 1;

      if (StartShift = FVisibleLinesCount-1) or ((Result <> 0)) then break;

      Inc(StartShift);
    end;
  end;

  function SearchScopeBackward: Integer;
  var
    l: Integer;
    uc: Integer;
    S: String;
  begin
    uc := FUpperBuffer.Count;
    Result := 0;

    while True do
    begin
      l := FLineIndex + StartShift;
      if l >= uc then
        S := FLowerBuffer[l-uc]
      else
        S := FUpperBuffer[l];

      if (FLastFoundCol < Length(S)) and (FLastFoundCol <> -1) then
        SetLength(S, FLastFoundCol);

      Result := RPos(Text, S);

      if (StartShift = 0) or (Result <> 0) then break;

      FLastFoundCol := -1;

      Dec(StartShift);
    end;
  end;

  function SearchScopeBackwardU: Integer;
  var
    l: Integer;
    uc: Integer;
    S: String;
  begin
    uc := FUpperBuffer.Count;
    Result := 0;

    while True do
    begin
      l := FLineIndex + StartShift;
      if l >= uc then
        S := AnsiUpperCase(FLowerBuffer[l-uc])
      else
        S := AnsiUpperCase(FUpperBuffer[l]);

      if (FLastFoundCol < Length(S)) and (FLastFoundCol <> -1) then
        SetLength(S, FLastFoundCol);

      Text := AnsiUpperCase(Text);
      
      Result := RPos(Text, S);

      if (StartShift = 0) or (Result <> 0) then break;

      FLastFoundCol := -1;

      Dec(StartShift);
    end;
  end;

begin
{
  SaveStartPos := FStartPos;
  SaveMidPos := FMidPos;
  SaveEndPos := FEndPos;
  SaveLineIndex := FLineIndex;
  SaveShift := FActiveLineShift;
  SavePos := FStream.Position;
}
  Result := False;

  if FStream = nil then exit;

  SaveCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;

  try

    UpdateActiveLine(True);
    FSilentMove := True;

    GetVisibleLinesCount;

    if vfoEntireScope in FindOptions then
    begin
      FLastFoundCol := -1;
      FLastFoundLine := -1;
      if vfoForward in FindOptions then
        InternalMoveTop
      else
        InternalMoveBottom;
    end
    else
      FLastFoundCol := FActivePosShift;

    StartShift := FActiveLineShift;
    FLastFoundLine := FGlobalLine + FActiveLineShift;

    while True do
    begin

      if vfoForward in FindOptions then
        if vfoIgnoreCase in Findoptions then
          WhereFoundSomething := SearchScopeForwardU
        else
          WhereFoundSomething := SearchScopeForward
      else
        if vfoIgnoreCase in Findoptions then
          WhereFoundSomething := SearchScopeBackwardU
        else
          WhereFoundSomething := SearchScopeBackward;

      if WhereFoundSomething <> 0 then
      begin
        FLastFoundCol := WhereFoundSomething;
        if vfoForward in FindOptions then
          FActivePosShift := WhereFoundSomething-1+Length(Text)
        else
          FActivePosShift := WhereFoundSomething-1;

        FLastFoundLine := FGlobalLine + StartShift;
        FActiveLineShift := StartShift;
        Result := True;

        if vfoForward in FindOptions then
          FMarkBegin := TextPos(FLastFoundLine, FActivePosShift - Length(Text))
        else
          FMarkBegin := TextPos(FLastFoundLine, FActivePosShift + Length(Text));
        FMarkEnd   := FMarkBegin;
        FMarkStartPoint := FMarkBegin;
        FMarking := True;
        FMarkType := mtHorizontal;
        break;
      end;

      Old := FLineIndex;

      if vfoForward in FindOptions then
      begin
        InternalMovePageDown;
        StartShift := 0;
      end
      else
      begin
        InternalMovePageUp;
        StartShift := FVisibleLinesCount - 1;
      end;

      if Old = FLineIndex then break;
    end;

    UpdateMarking;

    if FLineIndex + FActiveLineShift < FUpperBuffer.Count then
      S := FUpperBuffer.Strings[FLineIndex + FActiveLineShift]
    else
      S := FLowerBuffer.Strings[FLineIndex + FActiveLineShift - FUpperBuffer.Count];

    if CalcPosShift(S, FActivePosShift) + FHorzShift <= 0 then
    begin
      SetLength(S, FActivePosShift - Length(Text));
      FHorzShift := -CalcPosShift(S, Length(S));
    end
    else if CalcPosShift(S, FActivePosShift) + FHorzShift >= Self.Width then
      Dec(FHorzShift, CalcPosShift(S, FActivePosShift) + FHorzShift - Self.Width);

    Invalidate;
    FSilentMove := False;


    if voCaret in ViewOptions then
      UpdateCaret;

    FMarking := False;
  finally
    Screen.Cursor := SaveCursor;
  end;
end;

procedure TCustomView.InternalMoveLineHome;
begin
  CheckMarking;
  FActivePosShift := 0;
  if voCaret in ViewOptions then
  begin
    UpdateCaret;
    InternalMoveCaret(1, False);
    InternalMoveCaret(-1, False);
  end
  else
    ScrollText(-FHorzShift);
end;

procedure TCustomView.InternalMoveLineEnd;
var
  S: String;
begin
  if FStream = nil then exit;

  CheckMarking;
  if FLineIndex + FActiveLineShift >= FUpperBuffer.Count then
    S := FLowerBuffer[FLineIndex + FActiveLineShift - FUpperBuffer.Count]
  else
    S := FUpperBuffer[FLineIndex + FActiveLineShift];

  if voCaret in ViewOptions then
  begin
    FActivePosShift := Length(S);
    UpdateCaret;
    InternalMoveCaret(1, True);
    InternalMoveCaret(-1, True);
  end
  else
    ScrollText(-CalcPosShift(S, Length(S)) - FHorzShift + Self.Width);
end;

procedure TCustomView.StartMarking(MarkType: TMarkType);
begin

  if FMarking then exit;

  FMarkType := MarkType;
  FMarkBegin := TextPos(FGlobalLine+FActiveLineShift, FActivePosShift);
  FMarkEnd   := FMarkBegin;
  FMarkStartPoint := FMarkBegin;
  FMarking   := True;
  FGlobalMarkOffset := GetGlobalOffset(FMarkBegin.Line);
  Invalidate;
  Update;
end;

procedure TCustomView.UpdateMarking;
var
  tmp: TTextPos;
begin
  if not FMarking then exit;
  tmp := TextPos(FGlobalLine+FActiveLineShift, FActivePosShift);

  if LessThen(tmp, FMarkStartPoint) then
  begin
    FMarkBegin := tmp;
    FMarkEnd := FMarkStartPoint;
    if FMarkType = mtVertical then
    begin
      FMarkBegin.Col := FMarkEnd.Col;
      FMarkEnd.Col := tmp.Col;
    end;
    FGlobalMarkOffset := GetGlobalOffset(FMarkBegin.Line);
  end
  else
  begin
    FMarkBegin := FMarkStartPoint;
    FMarkEnd := tmp;
  end;

  if (FMarkType = mtVertical) and (FMarkBegin.Col > FMarkEnd.Col) then
  begin
    FMarkEnd.Col := FMarkBegin.Col;
    FMarkBegin.Col := tmp.Col;
  end;
end;

procedure TCustomView.FinishMarking(MarkType: TMarkType);
begin
  if (FMarkType <> MarkType) and (MarkType <> mtAny) then exit;

  FMarking := False;
  if (FMarkEnd.Line = FMarkBegin.Line) and
     (FMarkEnd.Col = FMarkBegin.Col) then
    CancelMarking;
end;

procedure TCustomView.CancelMarking;
begin
  if not SomethingMatched then exit;
  FMarkType := mtNone;
  Invalidate;
  Update;
end;

procedure TCustomView.CheckMarking;
begin
  if ((GetKeyState(VK_SHIFT) < 0) or
      (GetKeyState(VK_LBUTTON) < 0)) and (GetKeyState(VK_MENU) < 0) then
    StartMarking(mtVertical)
  else if (GetKeyState(VK_SHIFT) < 0) or (GetKeyState(VK_LBUTTON) < 0) then
    StartMarking(mtHorizontal)
  else
    if (not FMarking) and (not (voPersistentBlocks in ViewOptions)) then
      CancelMarking;
end;

procedure TCustomView.FreeClipboardHandle;
begin
  GlobalFree(FClipboardHandle);
end;

procedure TCustomView.WriteBlockToStream(S: TStream);
var
  i1, i2: Integer;
  SavePos: Integer;
  cnt: Integer;
  Str: String;
  P: PChar;
  EOL: array [0..1] of char;
  SaveCursor: Integer;
begin
  if Equel(FMarkBegin, FMarkEnd) then exit;

  with S do
  begin
    SavePos := FStream.Position;
    SaveCursor := Screen.Cursor;
    Screen.Cursor := crHourGlass;
    try
      FStream.Position := FGlobalMarkOffset;
      cnt := FMarkEnd.Line - FMarkBegin.Line + 1;

      EOL[0] := #13;
      EOL[1] := #10;

      while cnt > 0 do
      begin
        if not ReadLineForward(Str, lrsBlockWriting) then break;

        if (cnt = FMarkEnd.Line - FMarkBegin.Line + 1) or
           (FMarkType = mtVertical) then
          i1 :=  Min(Length(Str), FMarkBegin.Col)
        else
          i1 := 0;

        if (cnt = 1) or (FMarkType = mtVertical) then
          i2 := Min(Length(Str), FMarkEnd.Col)
        else
          i2 := Length(Str);

        P := PChar(Str);

        S.WriteBuffer(P[i1], i2-i1);
        if Cnt > 1 then
          S.WriteBuffer(EOL[0], 2);

        Dec(Cnt);
      end;
    finally
      FStream.Position := SavePos;
      Screen.Cursor := SaveCursor;
    end;
  end;
end;

procedure TCustomView.SaveBlock(FileName: String);
var
  S: TStream;
begin
  S := TFileStream.Create(FileName, fmCreate);
  try
    WriteBlockToStream(S);
  finally
    S.Free;
  end;
end;

procedure TCustomView.CopyToClipboard;
var
  MemStream: TMemoryStream;
  pData: Pointer;
  SaveCursor: Integer;
begin
  if FMarkType = mtNone then exit;

  SaveCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  MemStream := TMemoryStream.Create;
  try
    WriteBlockToStream(MemStream);
    Clipboard.Open;
    try
      Clipboard.Clear;

      FClipboardHandle := GlobalAlloc(GMEM_MOVEABLE, MemStream.Size+1);
      pData := GlobalLock(FClipboardHandle);
      try
        Move(MemStream.Memory^, pData^, MemStream.Size+1);
        Clipboard.SetAsHandle(CF_TEXT, FClipboardHandle);
      finally
        GlobalUnLock(FClipboardHandle);
      end;
    finally
      Clipboard.Close;
    end;
  finally
    MemStream.Free;
    Screen.Cursor := SaveCursor;
  end;
end;

procedure TCustomView.WMDestroyClipboard(var Msg: TMessage);
begin
  FreeClipboardHandle;
  inherited;
end;

function TCustomView.GetGlobalOffset(GlobalLine: Integer): Integer;
var
  i: Integer;
  l: Integer;
  uc: Integer;
begin
  i := 0;
  l := 0;
  uc := FUpperBuffer.Count;

  while (i + FGlobalLine - FLineIndex) < GlobalLine do
  begin
    if i < uc then
      Inc(l, Length(FUpperBuffer.Strings[i]))
    else
      Inc(l, Length(FLowerBuffer.Strings[i-uc]));
    Inc(l,2);
    Inc(i);
  end;

  Result := l + FStartPos;
end;

procedure TCustomView.ColorChanged;
begin
  UpdateActiveLine(True);
  if FSelectedColor <> nil then
    FActiveBrush.Color := FSelectedColor.BackColor;
  UpdateActiveLine(False);
end;

function TCustomView.CalcRealShift(LineShift: Integer; Shift: Integer): Cardinal;
var
  l: Integer;
  uc: Integer;
  P: PChar;
  ch: Char;
  w: Integer;
  OldFont: THandle;
begin
  Result := 0;
  uc := FUpperBuffer.Count;

  if (uc = 0) and (FLowerBuffer.Count = 0) then exit;

  l := FLineIndex + LineShift;

  if l < uc then
    P := PChar(FUpperBuffer.Strings[l])
  else
    P := PChar(FLowerBuffer.Strings[l-uc]);

  Dec(Shift, FHorzShift);

  l := 0;
  OldFont := SelectObject(ActiveDC, Font.Handle);
  try
    while l < Shift do
    begin
      if Result < StrLen(P) then
        ch := P[Result]
      else
        ch := ' ';

      GetCharWidth(ActiveDC, Ord(ch), Ord(ch), w);
      inc(l, w);
      Inc(Result);
    end;
    if Result < StrLen(P) then Dec(Result);
  finally
    SelectObject(ActiveDC, OldFont);
  end;
end;

procedure TCustomView.InternalPrint(Title: String; StartPage, EndPage: Integer; IgnorePageBreaks: Boolean);
var
  SavePos: Integer;
  S: String;
  PageNum: Integer;
  DeviceHeight: Integer;
  DeviceWidth: Integer;
  OldDC: HDC;
  Y: Integer;
  SaveCursor: Integer;
  PrintAction: TPrintAction;
begin
  if FStream = nil then exit;

  SaveCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;

  try
    Self.Update;
    if voPrintToDOSDevice in ViewOptions then
    begin
      DOSInternalPrint(StartPage, EndPage, IgnorePageBreaks);
      exit;
    end;

    PrintAction := paIgnoreCurrent;
    Printer.Title := Title;
    with Printer do
    begin
      Canvas.Font.Assign(Self.Font);
      if FPrintFontSize <> -1 then
        Canvas.Font.Size := FPrintFontSize;

      DeviceHeight := GetDeviceCaps(Printer.Canvas.Handle, VERTRES);
      DeviceWidth  := GetDeviceCaps(Printer.Canvas.Handle, HORZRES);

      PageNum := 0;

      OldDC := FActiveDC;
      FActiveDC := Canvas.Handle;
      RecalcMargins;
      FActiveDC := OldDC;

      Y := PageInfo.TopMargin;

      SavePos := FStream.Position;
      try
        FStream.Position := 0;
        while ReadLineForward(S, lrsWindowsPrinting) do
          if (Length(S)> 0) and (Pos(#12, S) > 0) then
          begin
            if (PageNum >= StartPage) and
               ((PageNum < EndPage) or (EndPage = -1)) and (not IgnorePageBreaks) then
              if (ptEntireRange = FPrintingType) or
                 ((ptOddRange = FPrintingType) and (PageNum mod 2 = 0)) or
                 ((ptEvenRange = FPrintingType) and (PageNum mod 2 <> 0)) then
                NewPage;

            Inc(PageNum);
            Y := PageInfo.TopMargin;
          end
          else
          begin
            if Y >= DeviceHeight then
            begin
              Y := PageInfo.TopMargin;
              if (PageNum >= StartPage) and
                 ((PageNum < EndPage) or (EndPage = -1)) and
                 (not IgnorePageBreaks) then
                if (ptEntireRange = FPrintingType) or
                   ((ptOddRange = FPrintingType) and (PageNum mod 2 = 0)) or
                   ((ptEvenRange = FPrintingType) and (PageNum mod 2 <> 0)) then
                Printer.NewPage;
              Inc(PageNum);
            end;

            if S = '' then S := ' ';

            if (PageNum >= StartPage) and
               ((PageNum <= EndPage) or (EndPage = -1)) then
             if (ptEntireRange = FPrintingType) or
                 ((ptOddRange = FPrintingType) and (PageNum mod 2 = 0)) or
                 ((ptEvenRange = FPrintingType) and (PageNum mod 2 <> 0)) then
                 if (PrintAction <> paIgnoreAll) and
                    (Canvas.TextWidth(S) + PageInfo.LeftMargin > DeviceWidth) then
                 begin
                   if Assigned(FOnPageOutOfWidth) then
                     FOnPageOutOfWidth(Self, PrintAction);

                   if PrintAction = paAbort then
                     SysUtils.Abort;
                 end
                 else
                   Canvas.TextOut(PageInfo.LeftMargin, Y, S);

            Inc(Y, Canvas.TextHeight(S));
            Inc(Y, PageInfo.LineGap);
            if (PageNum > EndPage) and (EndPage <> -1) then break;
          end;
         FStream.Position := SavePos;
      finally
        FStream.Position := SavePos;
      end;
    end
  finally
    Screen.Cursor := SaveCursor;
  end;
end;

procedure TCustomView.DOSInternalPrint(StartPage, EndPage: Integer; IgnorePageBreaks: Boolean);
var
  SavePos: Integer;
  S: String;
  PageNum: Integer;
  LineCount: Integer;
  F: TextFile;
begin

  Self.Update;

  AssignFile(F, FDOSPrintDevice);
  Rewrite(F);
  try
    PageNum := 0;
    LineCount := 0;
    SavePos := FStream.Position;
    try
      FStream.Position := 0;
      while ReadLineForward(S, lrsDOSPrinting) do
      begin
        if S = ''          then S := ' ';
        if Pos(#12, S) > 0 then S := #12;

        if (PageNum >= StartPage) and
           ((PageNum <= EndPage) or (EndPage = -1)) and
           ((S <> #12) or (not IgnorePageBreaks)) then
          if (ptEntireRange = FPrintingType) or
             ((ptOddRange = FPrintingType) and (PageNum mod 2 = 0)) or
             ((ptEvenRange = FPrintingType) and (PageNum mod 2 <> 0)) then
            WriteLn(F, S);
        Inc(LineCount);

        if (Pos(#12, S) > 0) or (LineCount >= FLinesPerPage) then
        begin
          LineCount := 0;
          Inc(PageNum);
        end;

        if (PageNum > EndPage) and (EndPage <> -1) then break;
      end;
    finally
      FStream.Position := SavePos;
    end;
  finally
    System.CloseFile(F);
  end
end;

procedure TCustomView.Print(Title: String; IgnorePageBreaks: Boolean);
begin
  if FStream = nil then exit;

  if voPrintToDOSDevice in ViewOptions then
  begin
    DOSInternalPrint(0, -1, IgnorePageBreaks);
    exit;
  end;

  Printer.BeginDoc;
  try
    InternalPrint(Title, 0, -1, IgnorePageBreaks);
    Printer.EndDoc;
  except
    Printer.Abort;
    raise;
  end;

end;

procedure TCustomView.DOSPrint;
var
  SavePos: Integer;
  S: String;
  F: TextFile;
begin
  AssignFile(F, FDOSPrintDevice);
  Rewrite(F);
  try
    SavePos := FStream.Position;
    FStream.Position := 0;
    try
      while ReadLineForward(S, lrsDOSPrinting) do
        WriteLn(F, S);
    finally
      FStream.Position := SavePos;
    end;
  finally
    System.CloseFile(F);
  end;
end;

procedure TCustomView.PrintBlock(Title: String; IgnorePageBreaks: Boolean);
var
  i1, i2: Integer;
  SavePos: Integer;
  cnt: Integer;
  Str: String;
  Y: Integer;
  SaveCursor: Integer;
begin
  if FStream = nil then exit;

  SaveCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;

  try
    Self.Update;
    if voPrintToDOSDevice in ViewOptions then
    begin
      DOSPrintBlock(IgnorePageBreaks);
      exit;
    end;
    Printer.Title := Title;
    with Printer do
    begin
      BeginDoc;
      Canvas.Font.Assign(Self.Font);
      if FPrintFontSize <> -1 then
      Canvas.Font.Size := FPrintFontSize;

      SavePos := FStream.Position;
      Y := 0;
      try
        FStream.Position := FGlobalMarkOffset;
        cnt := FMarkEnd.Line - FMarkBegin.Line + 1;

        while cnt > 0 do
        begin
        if not ReadLineForward(Str, lrsBlockWriting) then break;

          if (cnt = FMarkEnd.Line - FMarkBegin.Line + 1) or
             (FMarkType = mtVertical) then
            i1 :=  Min(Length(Str), FMarkBegin.Col+1)
          else
          i1 := 1;

          if (cnt = 1) or (FMarkType = mtVertical) then
            i2 := Min(Length(Str), FMarkEnd.Col+1)
          else
          i2 := Length(Str)+1;

          if (Length(Str)> 0) and (Pos(#12, Str) > 0) and
             (not IgnorePageBreaks) then
          begin
            NewPage;
            Y := 0;
          end
          else if i1 <= Length(Str) then
          begin
            Str := Copy(Str, i1, i2-i1);
            Canvas.TextOut(0, Y, Str);
            Inc(Y, Canvas.TextHeight(Str));
          end;

          Dec(Cnt);
        end;
      finally
        FStream.Position := SavePos;
        Printer.EndDoc;
      end;
    end;
  finally
    Screen.Cursor := SaveCursor;
  end;
end;

procedure TCustomView.DOSPrintBlock(IgnorePageBreaks: Boolean);
var
  SavePos: Integer;
  F: TextFile;
  Str: String;
  i1, i2: Integer;
  cnt: Integer;
  CanContinue: Boolean;
begin
  AssignFile(F, FDOSPrintDevice);
  Rewrite(F);
  SavePos := FStream.Position;
  FStream.Position := FGlobalMarkOffset;
  cnt := FMarkEnd.Line - FMarkBegin.Line + 1;
  try
    while cnt > 0 do
    begin
      if not ReadLineForward(Str, lrsDOSBlockWriting) then break;

      if (cnt = FMarkEnd.Line - FMarkBegin.Line + 1) or
         (FMarkType = mtVertical) then
        i1 :=  Min(Length(Str), FMarkBegin.Col+1)
      else
        i1 := 1;

      if (cnt = 1) or (FMarkType = mtVertical) then
        i2 := Min(Length(Str), FMarkEnd.Col+1)
      else
        i2 := Length(Str)+1;

      CanContinue := True;

      if (Length(Str)> 0) and
         (Pos(#12, Str) > 0) and
         IgnorePageBreaks then
        CanContinue := False;

      if CanContinue and (i1 <= Length(Str)) then
      begin
        Str := Copy(Str, i1, i2-i1);
        if Pos(#12, Str) > 0 then Str := #12;
        WriteLn(F, Str);
      end;

      Dec(Cnt);
    end;
  finally
    FStream.Position := SavePos;
    System.CloseFile(F);
  end;
end;

procedure TCustomView.PrintRange(Title: String; FromPage, ToPage: Integer; IgnorePageBreaks: Boolean);
begin
  Printer.BeginDoc;
  try
    InternalPrint(Title, FromPage-1, ToPage-1, IgnorePageBreaks);
    Printer.EndDoc;
  except;
    Printer.Abort;
  end;
end;

function TCustomView.SomethingMatched: Boolean;
begin
  Result := not Equel(FMarkBegin, FMarkEnd) and (FMarkType <> mtNone);
end;

type
  TBlockStatus = (bsNormal, bsBadSingle, bsBadFirst, bsBadLast, bsNone);

  TFetchBlock = record
    StartPage: Integer;
    EndPage: Integer;
    Status: TBlockStatus;
    NewPos: Integer;
  end;

procedure TCustomView.PrintFromString(Title: String; PrintString: String; IgnorePageBreaks: Boolean);

  function FetchBlock(CurPos: Integer): TFetchBlock;
  var
    S: String;
    WorkS: String;
    l: Integer;
  begin
    if CurPos >= Length(PrintString) then
    begin
      Result.Status := bsNone;
      exit;
    end;

    S := Copy(PrintString, CurPos + 1, Length(PrintString));
    l := Pos(',', S);

    if (l = 0) and (Length(S) > 0) then
      WorkS := S
    else
      WorkS := Copy(S, 1, l-1);

    Result.NewPos := CurPos + Length(WorkS) + 1;
    WorkS := Trim(WorkS);

    l := Pos('-', WorkS);
    if l = 0 then
      try
        Result.StartPage := StrToInt(WorkS);
        Result.EndPage := Result.StartPage;
        Result.Status := bsNormal;
      except
        Result.Status := bsBadSingle;
      end
    else
    begin
      S := Trim(Copy(WorkS, 1, l -1));
      try
        Result.StartPage := StrToInt(S);
        S := Trim(Copy(WorkS, l + 1, Length(WorkS)));
        try
          Result.EndPage := StrToInt(S);
          Result.Status := bsNormal;
        except
          Result.Status := bsBadLast;
        end;
      except;
        Result.Status := bsBadFirst;
      end;
    end;
  end;
var
  APos: Integer;
  FB: TFetchBlock;
begin
  Printer.BeginDoc;
  APos := 0;
  try
    while True do
    begin
      FB := FetchBlock(APos);
      if FB.Status = bsNone then break;

      if FB.Status = bsNormal then
      begin
        if (APos > 0) and (not IgnorePageBreaks) then
          Printer.NewPage;
        InternalPrint(Title, FB.StartPage-1, FB.EndPage-1, IgnorePageBreaks);
      end
      else
        raise ERangeError.Create('Error in range');

      APos := FB.NewPos;
    end;
    Printer.EndDoc;
  except
    Printer.Abort;
    raise;
  end;
end;

end.
