unit PrintSetup;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, Spin, WinSpool;

type
  TPrintingType = (ptEntireRange, ptOddRange, ptEvenRange);
  TPrintingTypes = set of TPrintingType;
  TPrintingRange = (prAll, prRange, prSelected);
  TPrinterName = array [0..79] of char;

  TPrintSetupDefaults = class(TPersistent)
  private
    FPaperSize: Short;
    FOrientation: Short;
    FXResolution: Short;
    FYResolution: Short;
  public
    constructor Create;
  published
    property PaperSize: Short read FPaperSize write FPaperSize;
    property Orientation: Short read FOrientation write FOrientation;
    property XResolution: Short read FXResolution write FXResolution;
    property YResolution: Short read FYResolution write FYResolution;
  end;

  TPrintSetupDialogEx = class(TForm)
    GroupBox1: TGroupBox;
    StaticText1: TStaticText;
    cbNames: TComboBox;
    btnProperties: TButton;
    StaticText2: TStaticText;
    StaticText3: TStaticText;
    StaticText4: TStaticText;
    StaticText5: TStaticText;
    stState: TStaticText;
    stType: TStaticText;
    stPort: TStaticText;
    stNotes: TStaticText;
    RadioGroup1: TRadioGroup;
    rbAll: TRadioButton;
    rbSelected: TRadioButton;
    rbPages: TRadioButton;
    efPages: TEdit;
    GroupBox2: TGroupBox;
    StaticText7: TStaticText;
    efCopies: TSpinEdit;
    cbCollate: TCheckBox;
    imSep: TImage;
    imMixed: TImage;
    GroupBox3: TGroupBox;
    StaticText8: TStaticText;
    StaticText9: TStaticText;
    efTopMargin: TEdit;
    efLeftMargin: TEdit;
    StaticText10: TStaticText;
    seFontSize: TSpinEdit;
    btnOk: TButton;
    btnCancel: TButton;
    cbPrintToLPT: TCheckBox;
    StaticText12: TStaticText;
    cbPrintingTypes: TComboBox;
    cbToOEM: TCheckBox;
    Bevel1: TBevel;
    StaticText6: TStaticText;
    StaticText11: TStaticText;
    efLinesPerPage: TEdit;
    stLines: TStaticText;
    StaticText13: TStaticText;
    efLineGap: TEdit;
    procedure cbCollateClick(Sender: TObject);
    procedure btnPropertiesClick(Sender: TObject);
    procedure btnOkClick(Sender: TObject);
    procedure cbNamesChange(Sender: TObject);
    procedure rbPagesClick(Sender: TObject);
    procedure efTopMarginKeyPress(Sender: TObject; var Key: Char);
    procedure efPagesKeyPress(Sender: TObject; var Key: Char);
    procedure efPagesChange(Sender: TObject);
    procedure efTopMarginChange(Sender: TObject);
    procedure cbPrintToLPTClick(Sender: TObject);
  private
    { Private declarations }
    FPrinterDefaults: TPrintSetupDefaults;

    FDevMode: THandle;
    FPPrinterInfo2: PPrinterInfo2;
    hPrinter: THandle;

    FDevice: TPrinterName;
    FDriver: TPrinterName;
    FPort: TPrinterName;

    FPrintingType: TPrintingType;
    FCollate: Boolean;
    FCopies: Short;
    FPrintingRange: TPrintingRange;
    FPrintToLPT: Boolean;
    FConvertToOEM: Boolean;
    FTopMargin: Integer;
    FLeftMargin: Integer;
    FFontSize: Integer;
    FLinesPerPage: Integer;
    FLineGap: Integer;
    procedure InitDialog;
    procedure FinalizeDialog;
    procedure CheckForPicture;
    procedure CheckForLines;
    procedure ApplyProperties;
    procedure ApplyToProperties;
    procedure ApplyToPrinter;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Execute: Boolean;
    procedure ApplyDefaults;

    property PrintingRange: TPrintingRange read FPrintingRange write FPrintingRange;
    property PrintingType: TPrintingType read FPrintingType write FPrintingType;
    property Collate: Boolean read FCollate write FCollate;
    property Copies: Short read FCopies write FCopies;
    property PrintToLPT: Boolean read FPrintToLPT write FPrintToLPT;
    property ConvertToOEM: Boolean read FConvertToOEM write FConvertToOEM;
    property FontSize: Integer read FFontSize write FFontSize;
    property TopMargin: Integer read FTopMargin write FTopMargin;
    property LeftMargin: Integer read FLeftMargin write FLeftMargin;
    property PrinterDefaults: TPrintSetupDefaults read FPrinterDefaults write FPrinterDefaults;
    property LinesPerPage: Integer read FLinesPerPage write FLinesPerPage;
    property LineGap: Integer read FLineGap write FLineGap;
  end;

  TPrintSetupDlg = class(TComponent)
  private
    FPrintSetupDialogEx: TPrintSetupDialogEx;
  protected
    function GetCollate: Boolean;
    procedure SetCollate(Value: Boolean);
    function GetCopies: Short;
    procedure SetCopies(Value: Short);
    function GetLeftMargin: Integer;
    procedure SetLeftMargin(Value: Integer);
    function GetTopMargin: Integer;
    procedure SetTopMargin(Value: Integer);
    function GetFontSize: Integer;
    procedure SetFontSize(Value: Integer);
    function GetPrintingRange: TPrintingRange;
    procedure SetPrintingRange(Value: TPrintingRange);
    function GetPrintingType: TPrintingType;
    procedure SetPrintingType(Value: TPrintingType);
    function GetPrintToLPT: Boolean;
    procedure SetPrintToLPT(Value: Boolean);
    function GetConvertToOEM: Boolean;
    procedure SetConvertToOEM(Value: Boolean);
    function GetCaption: String;
    procedure SetCaption(Value: String);
    function GetRange: String;
    procedure SetRange(Value: String);
    function GetPrinterDefaults: TPrintSetupDefaults;
    procedure SetPrinterDefaults(Value: TPrintSetupDefaults);
    function GetPortName: String;
    function GetDeviceName: String;
    function GetLinesPerPage: Integer;
    procedure SetLinesPerPage(Value: Integer);
    function GetLineGap: Integer;
    procedure SetLineGap(Value: Integer);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function Execute: Boolean;
    procedure ApplyDefaults;
    property PortName: String read GetPortName;
    property DeviceName: String read GetDeviceName;
  published
    property Collate: Boolean read GetCollate write SetCollate;
    property Copies: Short read GetCopies write SetCopies;
    property LeftMargin: Integer read GetLeftMargin write SetLeftMargin;
    property TopMargin: Integer read GetTopMargin write SetTopMargin;
    property FontSize: Integer read GetFontSize write SetFontSize;
    property PrintingRange: TPrintingRange read GetPrintingRange write SetPrintingRange;
    property PrintingType: TPrintingType read GetPrintingType write SetPrintingType;
    property PrintToLPT: Boolean read GetPrintToLPT write SetPrintToLPT;
    property ConvertToOEM: Boolean read GetConvertToOEM write SetConvertToOEM;
    property Caption: String read GetCaption write SetCaption;
    property Range: String read GetRange write SetRange;
    property PrinterDefaults: TPrintSetupDefaults read GetPrinterDefaults
                                                  write SetPrinterDefaults;
    property LinesPerPage: Integer read GetLinesPerPage write SetLinesPerPage;
    property LineGap: Integer read GetLineGap write SetLineGap;
  end;

procedure Register;

implementation

{$R *.DFM}
{$R *.DCR}

uses Printers;

var
  PrintingTypeText: array [TPrintingType] of string =
    ('  ',
     '  ',
     '  '
    );

procedure Register;
begin
  RegisterComponents('Dialogs', [TPrintSetupDlg]);
end;

function CopyData(Handle: THandle): THandle;
var
  Src, Dest: PChar;
  Size: Integer;
begin
  if Handle <> 0 then
  begin
    Size := GlobalSize(Handle);
    Result := GlobalAlloc(GHND, Size);
    if Result <> 0 then
      try
        Src := GlobalLock(Handle);
        Dest := GlobalLock(Result);
        if (Src <> nil) and (Dest <> nil) then Move(Src^, Dest^, Size);
      finally
        GlobalUnlock(Handle);
        GlobalUnlock(Result);
      end
  end
  else Result := 0;
end;

function PrintingTypeByText(Text: String): TPrintingType;
var
  pt: TPrintingType;
begin
  Result := ptEntireRange;
  for pt := Low(TPrintingType) to High(TPrintingType) do
    if PrintingTypeText[pt] = Text then
    begin
      Result := pt;
      break;
    end;
end;

{TPrintSetupDefaults}

constructor TPrintSetupDefaults.Create;
begin
  inherited Create;
  FPaperSize := -1;
  FOrientation := -1;
  FXResolution := -1;
  FYResolution := -1;
end;

{TPrintSetupDialogEx}

constructor TPrintSetupDialogEx.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDevMode := 0;
  FCollate := False;
  FCopies  := 1;
  FPrintingRange := prRange;
  FPrintToLPT := False;
  FPrintingType := ptEntireRange;
  FConvertToOEM := False;
  FTopMargin := 0;
  FLeftMargin := 0;
  FFontSize := 8;
  FLinesPerPage := 60;
  FPrinterDefaults := TPrintSetupDefaults.Create;
end;

destructor TPrintSetupDialogEx.Destroy;
begin
  if FPrinterDefaults <> nil then FPrinterDefaults.Free;
  inherited Destroy;
end;

function TPrintSetupDialogEx.Execute: Boolean;
begin
  FDevice := '';
  InitDialog;
  try
    Result := ShowModal = mrOk;
  finally
    FinalizeDialog;
  end;
end;

procedure TPrintSetupDialogEx.CheckForPicture;
begin
  imMixed.Visible := not cbCollate.Checked;
  imSep.Visible := cbCollate.Checked;
end;

procedure TPrintSetupDialogEx.CheckForLines;
begin
  efLinesPerPage.Enabled := cbPrintToLPT.Checked;
  stLines.Enabled := efLinesPerPage.Enabled;

  with efLinesPerPage do
    if Enabled then
      Color := clWindow
    else
      Color := clBtnFace;
end;

procedure TPrintSetupDialogEx.InitDialog;
var
  hDevMode: THandle;
  P: PDevMode;
  i: Integer;
  BadHandle: Boolean;
begin
  // initializing printers names
  cbNames.Items.Clear;

  with Printer do
    for i := 0 to Printers.Count-1 do
      cbNames.Items.Add(Printers[i]);

  Printer.GetPrinter(FDevice, FDriver, FPort, hDevMode);

  P := GlobalLock(hDevMode);
  try
    BadHandle := StrComp(FDevice, P^.dmDeviceName) <> 0;
  finally
    GlobalUnLock(hDevMode);
  end;

  if BadHandle then
  begin
    Printer.SetPrinter(FDevice, FDriver, FPort, 0);
    Printer.GetPrinter(FDevice, FDriver, FPort, hDevMode);
  end;

  // choose an active printer
  cbNames.ItemIndex := Printer.PrinterIndex;//cbNames.Items.IndexOf(FDevice);

  // fixing up DevMode for future

  // obtaining PRINTER_INFO_2
  OpenPrinter(FDevice, hPrinter, nil);
  GetPrinter(hPrinter, 2, nil, 0, @i);
  GetMem(FPPrinterInfo2, i);
  GetPrinter(hPrinter, 2, FPPrinterInfo2, i, @i);
  //
  with FPPrinterInfo2^ do
  begin
    stPort.Caption := pPortName;
    stType.Caption := pDriverName;
    stNotes.Caption := pComment;
  end;

  FDevMode := CopyData(hDevMode);

  P := GlobalLock(FDevMode);
  try
    if StrComp(FDevice, P^.dmDeviceName) <> 0 then
      DocumentProperties(0, hPrinter, FDevice, P^, P^, DM_IN_BUFFER or DM_OUT_BUFFER);

    with P^ do
    begin
      cbCollate.Checked := dmCollate = DMCOLLATE_TRUE;
      efCopies.Value := dmCopies;

      if PrinterDefaults.PaperSize <> -1 then
        dmPaperSize := PrinterDefaults.PaperSize;

      if PrinterDefaults.Orientation <> -1 then
        dmOrientation := PrinterDefaults.Orientation;

      if PrinterDefaults.XResolution <> -1 then
        dmPrintQuality := PrinterDefaults.XResolution;

      if PrinterDefaults.YResolution <> -1 then
        dmYResolution := PrinterDefaults.YResolution;

    end;
  finally
    GlobalUnLock(FDevMode);
  end;
  ApplyProperties;
end;

procedure TPrintSetupDialogEx.FinalizeDialog;
begin
  if FDevMode <> 0 then GlobalFree(FDevMode);
  if FPPrinterInfo2 <> nil then FreeMem(FPPrinterInfo2);
  ClosePrinter(hPrinter);

  FDevMode := 0;
end;

procedure TPrintSetupDialogEx.ApplyProperties;
var
  pt: TPrintingType;
begin
  with cbPrintingTypes.Items do
  begin
    Clear;
    for pt := Low(TPrintingType) to High(TPrintingType) do
      Add(PrintingTypeText[pt]);
    cbPrintingTypes.ItemIndex := IndexOf(PrintingTypeText[FPrintingType]);
  end;
  cbCollate.Checked := FCollate;
  efCopies.Value := FCopies;
  rbAll.Checked := FPrintingRange = prAll;
  rbSelected.Checked := FPrintingRange = prSelected;
  rbPages.Checked := FPrintingRange = prRange;

  cbPrintToLPT.Checked := FPrintToLPT;
  cbToOEM.Checked := FConvertToOEM;
  efTopMargin.Text := IntToStr(FTopMargin);
  efLeftMargin.Text := IntToStr(FLeftMargin);

  efLinesPerPage.Text := IntToStr(FLinesPerPage);
  efLineGap.Text := IntToStr(FLineGap);
  ActiveControl := cbNames;
  seFontSize.Value := Self.FontSize;
  CheckForPicture;
  CheckForLines;
end;

procedure TPrintSetupDialogEx.ApplyToProperties;
begin
  FCopies := efCopies.Value;
  FCollate := cbCollate.Checked;
  FLeftMargin := StrToInt(efLeftMargin.Text);
  FTopMargin  := StrToInt(efTopMargin.Text);
  FLineGap    := StrToInt(efLineGap.Text);

  FFontSize := seFontSize.Value;
  FPrintToLPT := cbPrintToLPT.Checked;
  FConvertToOEM := cbToOEM.Checked;
  FLinesPerPage := StrToInt(efLinesPerPage.Text);
  with cbPrintingTypes do
    FPrintingType := PrintingTypeByText(Items[ItemIndex]);

  if rbAll.Checked then
    FPrintingRange := prAll
  else if rbSelected.Checked then
    FPrintingRange := prSelected
  else
    FPrintingRange := prRange;
end;

procedure TPrintSetupDialogEx.ApplyToPrinter;
var
  sPrinterName: String;
  P: PDevMode;
begin
  with cbNames do
    sPrinterName := Items[ItemIndex];

  P := GlobalLock(FDevMode);
  try
    with P^ do
    begin
      dmCopies := efCopies.Value;
      if cbCollate.Checked then
        dmCollate := DMCOLLATE_TRUE
      else
        dmCollate := DMCOLLATE_FALSE;

      dmFields := DM_COPIES or DM_COLLATE or DM_ORIENTATION or
                  DM_ORIENTATION or
                  DM_PAPERSIZE or
                  DM_PAPERLENGTH or
                  DM_PAPERWIDTH	or
                  DM_SCALE or
                  DM_COPIES or
                  DM_DEFAULTSOURCE or
                  DM_PRINTQUALITY or
                  DM_COLOR or
                  DM_DUPLEX or
                  DM_YRESOLUTION or
                  DM_TTOPTION or
                  DM_COLLATE or
                  DM_FORMNAME or
                  DM_LOGPIXELS or
                  DM_BITSPERPEL	or
                  DM_PELSWIDTH or
                  DM_PELSHEIGHT	or
                  DM_DISPLAYFLAGS or
                  DM_DISPLAYFREQUENCY;
    end;
    Printer.SetPrinter(FDevice, FDriver, FPort, CopyData(FDevMode));
  finally
    GlobalUnLock(FDevMode);
  end;
end;

procedure TPrintSetupDialogEx.ApplyDefaults;
begin
  FDevice := '';
  InitDialog;
  try
    ApplyToPrinter;
  finally
    FinalizeDialog;
  end;
end;

procedure TPrintSetupDialogEx.cbCollateClick(Sender: TObject);
begin
  CheckForPicture;
end;

procedure TPrintSetupDialogEx.btnPropertiesClick(Sender: TObject);
var
  PrinterDefaults: TPrinterDefaults;
  P: PDevMode;
begin
  P := GlobalLock(FDevMode);
  try
    with PrinterDefaults do
    begin
      DesiredAccess := PRINTER_ACCESS_ADMINISTER;
      pDataType := '';
      pDevMode := PDeviceMode(P);
    end;

    DocumentProperties(Handle, hPrinter, FDevice, P^, P^, DM_IN_PROMPT or
                                                          DM_OUT_BUFFER or
                                                          DM_IN_BUFFER);
  finally
    GlobalUnLock(FDevMode);
  end;
end;

procedure TPrintSetupDialogEx.btnOkClick(Sender: TObject);
begin
  ApplyToProperties;
  ApplyToPrinter;
end;

procedure TPrintSetupDialogEx.cbNamesChange(Sender: TObject);
begin
  FinalizeDialog;
  Printer.PrinterIndex := cbNames.ItemIndex;
  InitDialog;
end;

procedure TPrintSetupDialogEx.rbPagesClick(Sender: TObject);
begin
  if rbPages.Checked and Visible then efPages.SetFocus;
end;

{TPrintSetupDlg}

constructor TPrintSetupDlg.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPrintSetupDialogEx := TPrintSetupDialogEx.Create(Self);
end;

destructor TPrintSetupDlg.Destroy;
begin
  if FPrintSetupDialogEx <> nil then
    FPrintSetupDialogEx.Free;
  inherited Destroy;
end;

function TPrintSetupDlg.GetCollate: Boolean;
begin
  Result := FPrintSetupDialogEx.Collate;
end;

function TPrintSetupDlg.GetCopies: Short;
begin
  Result := FPrintSetupDialogEx.Copies;
end;

function TPrintSetupDlg.GetLeftMargin: Integer;
begin
  Result := FPrintSetupDialogEx.LeftMargin;
end;

function TPrintSetupDlg.GetTopMargin: Integer;
begin
  Result := FPrintSetupDialogEx.TopMargin;
end;

function TPrintSetupDlg.GetFontSize: Integer;
begin
  Result := FPrintSetupDialogEx.FontSize;
end;

function TPrintSetupDlg.GetPrintingRange: TPrintingRange;
begin
  Result := FPrintSetupDialogEx.PrintingRange;
end;

function TPrintSetupDlg.GetPrintingType: TPrintingType;
begin
  Result := FPrintSetupDialogEx.PrintingType;
end;

function TPrintSetupDlg.GetPrintToLPT: Boolean;
begin
  Result := FPrintSetupDialogEx.PrintToLPT;
end;

function TPrintSetupDlg.GetConvertToOEM: Boolean;
begin
  Result := FPrintSetupDialogEx.ConvertToOEM;
end;

function TPrintSetupDlg.GetCaption: String;
begin
  Result := FPrintSetupDialogEx.Caption;
end;

procedure TPrintSetupDlg.SetCollate;
begin
  FPrintSetupDialogEx.Collate := Value;
end;

procedure TPrintSetupDlg.SetCopies;
begin
  FPrintSetupDialogEx.Copies := Value;
end;

procedure TPrintSetupDlg.SetLeftMargin;
begin
  FPrintSetupDialogEx.LeftMargin := Value;
end;

procedure TPrintSetupDlg.SetTopMargin;
begin
  FPrintSetupDialogEx.TopMargin := Value;
end;

procedure TPrintSetupDlg.SetFontSize;
begin
  FPrintSetupDialogEx.FontSize := Value;
end;

procedure TPrintSetupDlg.SetPrintingRange;
begin
  FPrintSetupDialogEx.PrintingRange := Value;
end;

procedure TPrintSetupDlg.SetPrintingType;
begin
  FPrintSetupDialogEx.PrintingType := Value;
end;

procedure TPrintSetupDlg.SetPrintToLPT;
begin
  FPrintSetupDialogEx.PrintToLPT := Value;
end;

procedure TPrintSetupDlg.SetConvertToOEM;
begin
  FPrintSetupDialogEx.ConvertToOEM := Value;
end;

procedure TPrintSetupDlg.SetCaption;
begin
  FPrintSetupDialogEx.Caption := Value;
end;

function TPrintSetupDlg.Execute: Boolean;
begin
  Result := FPrintSetupDialogEx.Execute;
end;

function TPrintSetupDlg.GetRange;
begin
  Result := FPrintSetupDialogEx.efPages.Text;
end;

procedure TPrintSetupDlg.SetRange;
begin
  FPrintSetupDialogEx.efPages.Text := Value;
end;

function TPrintSetupDlg.GetPrinterDefaults;
begin
  Result := FPrintSetupDialogEx.PrinterDefaults;
end;

procedure TPrintSetupDlg.SetPrinterDefaults;
begin
  FPrintSetupDialogEx.PrinterDefaults := Value;
end;

function TPrintSetupDlg.GetPortName: String;
begin
  Result := FPrintSetupDialogEx.stPort.Caption;
end;

function TPrintSetupDlg.GetDeviceName: String;
begin
  Result := FPrintSetupDialogEx.FDevice;
end;

function TPrintSetupDlg.GetLinesPerPage: Integer;
begin
  Result := FPrintSetupDialogEx.LinesPerPage;
end;

procedure TPrintSetupDlg.SetLinesPerPage(Value: Integer);
begin
  FPrintSetupDialogEx.LinesPerPage := Value;
end;

function TPrintSetupDlg.GetLineGap: Integer;
begin
  Result := FPrintSetupDialogEx.LineGap;
end;

procedure TPrintSetupDlg.SetLineGap(Value: Integer);
begin
  FPrintSetupDialogEx.LineGap := Value;
end;

procedure TPrintSetupDlg.ApplyDefaults;
begin
  FPrintSetupDialogEx.ApplyDefaults;
end;

procedure TPrintSetupDialogEx.efTopMarginKeyPress(Sender: TObject;
  var Key: Char);
begin
  if not (Key in ['0'..'9',',', #8]) then Key := #0;
end;

procedure TPrintSetupDialogEx.efPagesKeyPress(Sender: TObject;
  var Key: Char);
begin
  if not (Key in ['0'..'9',',', '-', #8]) then Key := #0;
end;

procedure TPrintSetupDialogEx.efPagesChange(Sender: TObject);
begin
  if rbPages.Enabled then rbPages.Checked := True;
end;

procedure TPrintSetupDialogEx.efTopMarginChange(Sender: TObject);
begin
  with TEdit(Sender) do
    if Text = '' then
    begin
      Text := '0';
      SelectAll;
    end;
end;

procedure TPrintSetupDialogEx.cbPrintToLPTClick(Sender: TObject);
begin
  CheckForLines;
end;

end.
