{**
*
*         Script Reporter Component library
*         TRepControl descendant and supplemental classes
*          Pavel Kilovaty 1998-99
*
*
**}

unit RepCtrl;

interface

uses SysUtils, Classes, RepKrnl, ObjList, Dialogs, RepExpr, Mask,
     StringTools, Reporter;

type
  TAlign=(taLeft,taCenter,taRight);

type
  {   Control}
  TRepExprCtrl = class
  private
    FExpression: TRepExpression;  //  expr
    FObjects: TObjectList;        //      . expr
    function GetValue(Index: Integer): Integer;
    function GetCount: Integer;
  public
    property Count: Integer read GetCount;
    property Values[Index: Integer]: Integer read GetValue;
    property Expression: TRepExpression read FExpression;
    constructor CreateOriginal(aExpression: TRepExpression);
    destructor Destroy;override;
    procedure Add(Value: Integer);
  end;

  {  " control"}
  TExprLine = class
  private
    FExprCount: Integer;             //  expr  
    FIndex: Integer;                 //   expr    FPosExpr
    FHeight: Integer;                //     
    FLastText: String;               //    expression  
    FLengthText: Integer;            //    (for optimization)
    procedure SetExprCount(Value: Integer);
    procedure SetHeight(Value: Integer);
    procedure SetLastText(Value: String);
    procedure SetIndex(Value: Integer);
    procedure SetLengthText(Value: Integer);
  public
    property LengthText: Integer read FLengthText write SetLengthText;
    property Index: Integer read FIndex write SetIndex;
    property LastText: String read FLastText write SetLastText;
    property ExprCount: Integer read FExprCount write SetExprCount;
    property Height: Integer read FHeight write SetHeight;
    constructor Create(aExprCount: Integer;aHeight: Integer);
  end;

  {   expression}
  TExprStorage = class
  private
    FAlign: TAlign;              //  -  
    FLine: Integer;              // N     expr
    FWidth: Integer;             //  
    FPrecision: Integer;         //  ( Float)
    FPicture: String;            //  
    FIndex: Integer;             //    expr
    FValue: TStringList;         //    
    FText: String;               // ,   expr
    FRepControl: TRepControl;    //  Reporter
    FTrim: Boolean;              //     
    FWordWrap: Boolean;          //    
    FUseFormat: Boolean;         //       
    procedure SetPicture(Value: String);
    procedure SetWordWrap(Value: Boolean);
    procedure SetAlign(Value: TAlign);
  protected
    function FormatWordWrap(Value: String;Width: Integer): String;
    {Added by Vlad 26.02.99}
    procedure FormatError(ErrorCode: Integer);
  public
    property Align: TAlign read FAlign write SetAlign;
    property Line: Integer read FLine;
    property Width: Integer read FWidth;
    property Precision: Integer read FPrecision;
    property Text: String read FText;
    property Picture: String read FPicture write SetPicture;
    property IndexExpr:Integer read FIndex;
    property WordWrap: Boolean read FWordWrap write SetWordWrap;
    property AllTrim: Boolean read FTrim;
    property UsePicture: Boolean read FUseFormat;
    constructor CreateOriginal(aLine: Integer;
                               aPicture: String;
                               aText: String;
                               aIndex: Integer;
                               RepControl: TRepControl;
                               aUsePicture: Boolean);
    destructor Destroy;override;
  end;

  { Control c exressions}
  TExprRepControl = class (TRepControl)
  private
    FFirstStep: Boolean;                  // for optimization
    FHeight: Integer;                     //  Control   
    FExpr: TObjectList;                   //  expr  Control (uniq)
    FPosExpr: TObjectList;                //   () expr  Control
    FLines: TStringList;                  // ,  Control
    FOptimized: Boolean;                  // for optimization
    procedure SetCurrExprValues;
    procedure SetCurrExprValuesOptimized;
    procedure SetDefaultPicture;
    function GetExpr(Index: Integer): TRepExpression;
  protected
    procedure Prepare; override;          // Compile control
    function GetHeight: Integer;override;
    procedure Turn;override;
  public
    constructor Create(RepEngine: TCustomRepEngine;
                WhenClause: TRepToken; AName: String; aLines: TStringList);
    destructor Destroy;override;
    property Expressions[Index: Integer]: TRepExpression read GetExpr;
    property Lines: TStringList read FLines;
  end;

implementation

procedure TExprStorage.SetWordWrap(Value: Boolean);
begin
  FWordWrap := Value;
end;

{TRepExprCtrl}

constructor TRepExprCtrl.CreateOriginal(aExpression: TRepExpression);
begin
  inherited Create;
  FExpression := aExpression;
  FObjects := TObjectList.Create;
end;

destructor TRepExprCtrl.Destroy;
begin
  FObjects.Free;
  inherited;
end;

procedure TRepExprCtrl.Add(Value: Integer);
begin
  FObjects.Add(TObject(Value));
end;

function TRepExprCtrl.GetValue(Index: Integer): Integer;
begin
  Result := Integer(FObjects.Items[Index]);
end;

function TRepExprCtrl.GetCount;
begin
  Result := FObjects.Count;
end;

{TExprLine}

constructor TExprLine.Create(aExprCount: Integer;aHeight: Integer);
begin
  inherited Create;
  FExprCount := aExprCount;
  FHeight := aHeight;
  FLastText := '';
  FLengthText := 0;
end;

procedure TExprLine.SetExprCount(Value: Integer);
begin
  FExprCount := Value;
end;

procedure TExprLine.SetLastText(Value: String);
begin
  FLastText := Value;
end;

procedure TExprLine.SetHeight(Value: Integer);
begin
  FHeight := Value;
end;

procedure TExprLine.SetIndex(Value: Integer);
begin
  FIndex := Value;
end;

procedure TExprLine.SetLengthText(Value: Integer);
begin
  FLengthText := Value;
end;

{TExprStorage}

constructor TExprStorage.CreateOriginal(aLine: Integer;
                                        aPicture: String;
                                        aText: String;
                                        aIndex: Integer;
                                        RepControl: TRepControl;
                                        aUsePicture: Boolean);
begin
  inherited Create;
  FLine := aLine;
  FIndex := aIndex;
  FText := aText;
  FRepControl := RepControl;
  FWordWrap := False;
  FTrim := False;
  FAlign := taLeft;
  Picture := aPicture;
  FValue := TStringList.Create;
  FUseFormat := aUsePicture;
end;

destructor TExprStorage.Destroy;
begin
  FValue.Free;
  inherited;
end;

procedure TExprStorage.SetAlign(Value: TAlign);
begin
  FAlign := Value;
end;

procedure TExprStorage.SetPicture(Value: String);
var
  PosDec, PosSep: Integer;
begin
  FPicture := Value;
  FWidth := length(FPicture);
  if Pos('@',FPicture) > 0  then
  begin
    Align := taCenter;
    FPicture := copy(FPicture,2,FWidth);
  end;

  if Pos('~',FPicture) > 0 then
  begin
    Value := trim(Value);
    FWordWrap := True;
    FPicture := copy(FPicture,2,FWidth);
    FWidth := StrToInt(FPicture);
  end;

  PosSep := Pos(';',FPicture);

  if PosSep > 0 then
    FWidth := PosSep-1
  else
    FWidth := length(FPicture);

  if (FWidth > 0)then
  begin
    if (FPicture[1] = '^') then
    begin
      FPicture := copy(FPicture,2,FWidth);
      FTrim := True;
    end;

    if AnsiUpperCase(FPicture[1])='W' then
    begin
      FPicture := copy(FPicture,2,FWidth);
      FWidth := FWidth-1;

      {Added by Vlad 26.02.99}
      if FWidth = 0 then FormatError(72);

      if (FPicture[1]>='0') and (FPicture[1]<='9') then
      begin
        FWidth := StrToInt(FPicture);
        FPicture := ReplicateChar('A',FWidth);
      end
      else // Added by Vlad 26.02.99
        FormatError(72);
    end;
  end;

  PosDec := Pos('.',FPicture);
  if  PosDec > 0 then
    FPrecision := length(trim(copy(FPicture,PosDec+1,PosSep-PosDec-1)))
  else
    FPrecision := 0;
end;

function TExprStorage.FormatWordWrap(Value: String;Width: Integer): String;
var
  s: String;
  ExistWord: Boolean;
  SpacePosition: Integer;
  NextLine: Integer;
  CurrLine: String;
begin
  s := '';
  ExistWord := True;
  CurrLine :='';
  Value := TrimRight(Value);
  Value := Value+' ';

  while ExistWord do
  begin
    SpacePosition := Pos(' ',Value);
    NextLine := SpacePosition;
    if NextLine < Width then
    begin
      while NextLine <= Width do
      begin
        if (SpacePosition > 0)and(SpacePosition<Width) then
        begin
          CurrLine := CurrLine+copy(Value,1,SpacePosition);
          Value:=copy(Value,SpacePosition+1,length(Value));
          SpacePosition := Pos(' ',Value);
          NextLine := NextLine + SpacePosition;
          if SpacePosition = 0 then
            NextLine := Width+2;
        end
        else
        begin
          CurrLine := CurrLine+copy(Value,1,Width);
          NextLine := Width + 2;
          ExistWord := False;
        end;
      end;
      if ExistWord then
      begin
        CurrLine := Padr(CurrLine,Width);
        s := s+CurrLine+chr(13);
        CurrLine := '';
      end;
    end
    else
    begin
      CurrLine := CurrLine+copy(Value,1,Width);
      if length(Value)<Width then
      begin
        ExistWord := False;
        CurrLine := Padr(CurrLine,Width);
        s := s+CurrLine+chr(13)+copy(Value,1,Width);
        CurrLine:='';
      end
      else
      begin
        Value := copy(Value,SpacePosition,length(Value));
        s := s+CurrLine+chr(13);
        CurrLine:='';
      end;
    end;
  end;

  Result := s;
end;

{ Added by Vlad 26.02.99}

procedure TExprStorage.FormatError(ErrorCode: Integer);
begin
  FRepControl.RepEngine.ReportError(72, FRepControl.Name, '',
    TRepExprCtrl(TExprRepControl(FRepControl).FExpr.Items[FIndex]).Expression,
    True);
end;

{TExprRepControl}

constructor TExprRepControl.Create(RepEngine: TCustomRepEngine;
                                   WhenClause: TRepToken;
                                   AName: String; aLines: TStringList);
var
  i: Integer;
  VarSL: TStringList;
  aText: PChar;
begin
  inherited CreateOriginal(RepEngine, WhenClause, AName);
  VarSL:= TStringList.Create;
  try
    aText:= '';
    for i:=0 to aLines.Count-1 do
      aText := PChar(String(aText)+aLines.Strings[i]);

    FOptimized := True;
    FLines := TStringList.Create;
    VarSL.SetText(aText);
    for i:=0 to VarSL.Count-1 do
      FLines.AddObject(VarSL.Strings[i],TExprLine.Create(0,0));
    FExpr := TObjectList.Create;
    FPosExpr := TObjectList.Create;
    FFirstStep := True;
  finally
    VarSL.Free;
  end;
end;

destructor TExprRepControl.Destroy;
var
  i: Integer;
begin

  for i:=0 to FLines.Count-1 do
    TExprLine(FLines.Objects[i]).Free;

  FLines.Free;
  FPosExpr.Free;
  FExpr.Clear;
  FExpr.Free;
  inherited Destroy;
end;

function TExprRepControl.GetExpr(Index: Integer): TRepExpression;
begin
  Result := FExpr[Index];
end;

{ Compile Control }

procedure TExprRepControl.Prepare;
var
  i,aExprCount, aLenExpr, aLenExprName: Integer;
  s: String;
  AllExpr: TStringList;
  ExprName,ExprString,aText,Picture: String;
  RepToken: TRepToken;
  Expr: TRepExpression;
  aLengthText: Integer;
  aUsePicture: Boolean;
begin
  AllExpr := nil;
  try
    AllExpr := TStringList.Create;
    try
      with FLines do
        for i:=0 to Count-1 do
        begin
          aLengthText := 0;
          aExprCount := 0;
          s:= Strings[i];
          while Pos('&',s)>0 do
          begin
            s:= trim(copy(s,Pos('&',s)+1,length(s)));

            //s:= copy(s,2,length(s)); - update by Pasha

            //    "expr_name,addr PICTURE expr_picture"
            ExprString := copy(s,1,Pos('&',s)-1);

            //   expr   s
            s := copy(s,Pos('&',s)+1,length(s));
            
            // for optimization - update
            aLenExpr := length(ExprString);

            aUsePicture := True;

            if Pos(' ',ExprString)>0 then
              ExprName := copy(ExprString,1,Pos(' ',ExprString))
            else
              ExprName := copy(ExprString,1,aLenExpr);

            ExprName := trim(ExprName);

            // for optimization - update
            aLenExprName := length(ExprName);

            RepToken := Str2RepToken(ExprName);

            if Pos('PICTURE',AnsiUpperCase(ExprString))>0 then
              Picture := copy(ExprString,Pos(' ',ExprString)+9,aLenExpr)
            else
            begin
              Picture := trim(copy(ExprString,
                              aLenExprName+1,
                              aLenExpr));
              if Picture <> '' then
                RepEngine.ReportError(15, RepToken.Text, Self.Name, Self, True)
              else
                aUsePicture := False;
            end;

            if Pos('NONE',AnsiUpperCase(ExprString)) > 0 then
              aUsePicture := False;
              
            aText := '';
            aText := copy(Strings[i],
                         1,
                         Pos('&'+ExprName,Strings[i])-1);
                         
            aLengthText := aLengthText+length(aText);

            { Updated by Vlad}
            if AllExpr.IndexOf(RepToken.Text)=-1 then
            begin
              if RepToken.Addr = nil then
                Expr := RepEngine.ExprByToken[RepToken]
              else
                Expr := RepToken.Addr;

              FExpr.Add( TRepExprCtrl.CreateOriginal(Expr) );

              AllExpr.Add(RepToken.Text);

              FPosExpr.Add(TExprStorage.CreateOriginal(
                                        i,Picture,
                                        aText,
                                        AllExpr.IndexOf(RepToken.Text),
                                        Self,aUsePicture));
              TRepExprCtrl(
              FExpr.Items[FExpr.Count-1]).Add(FPosExpr.Count-1);
            end
            else
            begin
              FOptimized := False;
              FPosExpr.Add(TExprStorage.CreateOriginal(
                                        i,Picture,
                                        aText,
                                        AllExpr.IndexOf(RepToken.Text),
                                        Self,aUsePicture));
              TRepExprCtrl(
              FExpr.Items[AllExpr.IndexOf(RepToken.Text)]).Add(FPosExpr.Count-1);
            end;
            
            
            Strings[i] := copy(Strings[i],
                               Pos(ExprName,Strings[i])+aLenExprName,
                               length(Strings[i]));
            Strings[i] := copy(Strings[i],Pos('&',Strings[i])+1,length(Strings[i]));
            if aExprCount = 0 then
              TExprLine(FLines.Objects[i]).Index := FPosExpr.Count-1;
            aExprCount := aExprCount+1;

          end;

          with TExprLine(FLines.Objects[i]) do
          begin
            LastText :=s;
            ExprCount := aExprCount;
          end;
          TExprLine(FLines.Objects[i]).LengthText := aLengthtext;
        end;
    except
      FExpr.Clear;
      raise;
    end;
  finally
    AllExpr.Free;
    inherited Prepare;
  end;
end;

{     }
procedure TExprRepControl.SetDefaultPicture;
var
  i,j,Index: Integer;
  aValue: Variant;
begin
  for i:=0 to FExpr.Count-1 do
  begin
    aValue := TRepExprCtrl(FExpr.Items[i]).Expression.Value;
    for j:=0 to TRepExprCtrl(FExpr.Items[i]).Count-1 do
    begin
        Index := TRepExprCtrl(FExpr.Items[i]).Values[j];
        with TExprStorage(FPosExpr.Items[Index]) do
         if Picture = '' then
           case VarType(aValue) of
             varSmallint,
             varInteger: Picture := '9999999999';
             varCurrency,
             varSingle,
             varDouble: Picture := '########0.00';
             varDate: Picture := 'c';
             varString, varOleStr: Picture := 'cccccccccc';
           end;
    end;
  end;
end;

{ ,   Control   }

procedure TExprRepControl.SetCurrExprValues;
var
  i,j: Integer;
  aValue: Variant;
  aString,s: String;
  Index: Integer;
begin
  for i:=0 to FExpr.Count-1 do
  begin
    aValue := TRepExprCtrl(FExpr.Items[i]).Expression.Value;

    for j:=0 to TRepExprCtrl(FExpr.Items[i]).Count-1 do
    begin
      Index := TRepExprCtrl(FExpr.Items[i]).Values[j];
      with TExprStorage(FPosExpr.Items[Index]) do
      begin
        if UsePicture then
          case VarType(aValue) of
             varSmallint,
             varInteger,
             varCurrency,
             varSingle,
             varDouble: begin
                          s := FormatFloat(Picture,aValue);
                          if s = '' then s := ' ';
                          s := s+Space(Precision-
                                 length(copy(s,Pos(DecimalSeparator,s)+1,20)));
                          aString := Padl(s,Width);
                        end;
             varDate: aString := FormatDateTime(Picture,aValue);
             varBoolean: aString := String(aValue);
             varString,
             varOleStr: begin
                          aString := String(aValue);
                          if WordWrap then
                            aString := FormatWordWrap(aString,Width)
                          else
                          begin
                            aString := FormatMaskText(Picture,aString);
                            if AllTrim then aString := trim(aString);
                            if Align = taCenter then aString := CenterStr(trim(aString),Width);
                          end;
                        end;
           else
             aString := Space(Width);
           end
         else
           aString := String(aValue);

         FValue.SetText(PChar(aString));
      end;
    end;
  end;
end;

{,   Control   }
procedure TExprRepControl.SetCurrExprValuesOptimized;
var
  i: Integer;
  aValue: Variant;
  s: String;
  aString: String;
begin

  for i:=0 to FExpr.Count-1 do
    with TExprStorage(FPosExpr.Items[i]) do
    begin
      aValue := TRepExprCtrl(FExpr.Items[i]).Expression.Value;
      if UsePicture then
        case VarType(aValue) of
             varSmallint,
             varInteger,
             varCurrency,
             varSingle,
             varDouble:begin
                         s := FormatFloat(Picture,aValue);
                         if s = '' then s := ' ';
                         s := s+Space(Precision-
                              length(copy(s,Pos(DecimalSeparator,s)+1,20)));
                         aString := Padl(s,Width);
                       end;
             varDate: aString := FormatDateTime(Picture,aValue);
             varBoolean: aString := String(aValue);
             varString,
             varOleStr: begin
                          aString := String(aValue);
                          if WordWrap then
                            aString := FormatWordWrap(aString,Width)
                          else
                          begin
                            aString := FormatMaskText(Picture,aString);
                            if AllTrim then aString := trim(aString);
                            if Align = taCenter then aString := CenterStr(trim(aString),Width);
                          end;
                        end;
        else
          aString := Space(Width);
        end
      else if aValue = Null then
           begin
             RepEngine.ReportError(16, TRepExprCtrl(FExpr.Items[i]).Expression.Name,
                                   Self.Name, TRepExprCtrl(FExpr.Items[i]).Expression, True);
           end
           else
             aString := String(aValue);

      FValue.SetText(PChar(aString));
    end;
end;

{   Control}

function TExprRepControl.GetHeight;
var
  i,j,l: Integer;
  MaxHeight,CurrHeight: Integer;
  aExprCount: Integer;
  CalcHeight: Boolean;
begin
  CalcHeight := False;
  MaxHeight := 1;
  CurrHeight := 0;
  //         
  if not Processed then
  begin
    CalcHeight := True;
    l := 0;
    //    expr   
    if FFirstStep then
    begin
      SetDefaultPicture;
      FFirstStep := False;
    end;

    if not FOptimized then
      SetCurrExprValues
    else
      SetCurrExprValuesOptimized;

    //    
    for i:= 0 to FLines.Count-1 do
    begin
      aExprCount := TExprLine(FLines.Objects[i]).ExprCount;
      //    expr  
      if aExprCount > 0 then
        with TExprStorage(FPosExpr.Items[l]) do
        begin
          for j:=0 to aExprCount-1 do
            if MaxHeight < FValue.Count then MaxHeight := FValue.Count;
          Inc(l);
        end;
      TExprLine(FLines.Objects[i]).Height := MaxHeight;
      CurrHeight := CurrHeight + MaxHeight;
      MaxHeight := 1;
    end;
  end;

  if not CalcHeight then CurrHeight := FHeight;

  FHeight := CurrHeight;
  Result := CurrHeight;

end;

{  Control   }

procedure TExprRepControl.Turn;
var
  i,j,l,m: Integer;
  OutString: string;
  NewLine: Boolean;
  CurrLineHeight: Integer;
  CurrIndexExpr,aExprCount: Integer;
  LeadingSpace: String;
begin
  l:=0;
  //    

  for i:= 0 to FLines.Count-1 do
  begin
    NewLine := True;
    with TExprLine(FLines.Objects[i]) do
    begin
      CurrLineHeight := Height;
      aExprCount := ExprCount;
    end;
      
    if aExprCount=0 then
      Print(FLines.Strings[i], nil)
    else
    begin
      //  max  expr   
      CurrIndexExpr := l;
      for j:=0 to CurrLineHeight-1 do
      begin
        OutString := '';
        //        
        if j>0 then l:= CurrIndexExpr;
        LeadingSpace := '';
        //   - expr  
        for m:=0 to aExprCount-1 do
        begin
          //      expr
          with TExprStorage(FPosExpr.Items[l]) do
            if FValue.Count-1 >= j then
              if NewLine then
              begin
                if m=aExprCount-1 then
                  OutString := OutString+Text+FValue.Strings[j]+
                               TExprLine(FLines.Objects[i]).LastText
                else
                  OutString := OutString+Text+FValue.Strings[j];
              end
              else
              begin
                LeadingSpace := LeadingSpace+Space(length(Text));
                OutString := OutString+LeadingSpace+FValue.Strings[j];
                LeadingSpace := '';
              end
            else if not NewLine then
                 LeadingSpace := LeadingSpace+
                            Space(length(Text)+Width);
            Inc(l);
          end;
          NewLine := False;
          Print(OutString, nil);
      end;
    end;
  end;
end;

end.

