{**
*
*         Script Reporter Component library
*         Report parameters handling example
*          Vlad Ilushchenko 1998-99
*
*
**}

unit RepParamForm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, RepKrnl, StdCtrls;

type
  TfmRepParam = class(TForm)
    sgData: TStringGrid;
    btnOk: TButton;
    btnCancel: TButton;
    procedure sgDataSetEditText(Sender: TObject; ACol, ARow: Integer;
      const Value: String);
    procedure sgDataGetEditMask(Sender: TObject; ACol, ARow: Integer;
      var Value: String);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    { Private declarations }
    FParams: TRepParams;
    FOldParams: TRepParams;
    FReporter: TCustomRepEngine;
    procedure AnalyzeParams;
    procedure FillParams;
  public
    { Public declarations }
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;
    function OpenForm(Reporter: TCustomRepEngine): Integer;

  end;

procedure AdjustSGSizes(SG: TStringGrid);

var
  fmRepParam: TfmRepParam;

implementation

{$R *.DFM}


procedure AdjustSGSizes(SG: TStringGrid);
var
  DC: HDC;
  OldFont: HFont;
  tm: TTextMetric;
begin
  with SG do
  begin
    DC := GetDC(Handle);
    try
      OldFont := SelectObject(DC, Font.Handle);
      try
        GetTextMetrics(DC, tm);
        DefaultRowHeight := tm.tmHeight + tm.tmExternalLeading + 4;
        DefaultColWidth := Width div 2;
      finally
        SelectObject(DC, OldFont);
      end;
    finally
      ReleaseDC(Handle, DC);
    end;
  end;
end;

constructor TfmRepParam.Create(Owner: TComponent);
begin
  inherited Create(Owner);
  FParams := TRepParams.Create(nil);
  FOldParams := TRepParams.Create(nil);
end;

destructor TfmRepParam.Destroy;
begin
  if FParams <> nil then FParams.Free;
  if FOldParams <> nil then FOldParams.Free;
  inherited Destroy;
end;

procedure TfmRepParam.AnalyzeParams;
var
  i: Integer;
begin
  FParams.Assign(FReporter.Params);
  FOldParams.Assign(FReporter.Params);
  FReporter.GetParams;
  if FParams.ParamCount <> FReporter.Params.ParamCount then
    FParams.Assign(FReporter.Params)
  else
    for i := 0 to FParams.ParamCount-1 do
      if FParams.Param[i].ParamName <> FReporter.Params.Param[i].ParamName then
        FParams.Assign(FReporter.Params);
end;

procedure TfmRepParam.FillParams;
var
  i: Integer;
begin
  with sgData do
  begin
    if FParams.ParamCount = 0 then
    begin
      RowCount := 1;
      Options := Options - [goEditing, goAlwaysShowEditor];
      Cells[0, 0] := '';
      Cells[1, 0] := '';
    end
    else
    begin
      RowCount := FParams.ParamCount+1;
      FixedRows := 1;
      Options := Options + [goEditing, goAlwaysShowEditor];
      Cells[0, 0] := 'Parameter';
      Cells[1, 0] := 'Value';
    end;

    for i := 0 to FParams.ParamCount-1 do
    begin
      Cells[0, i + 1] := FParams.Param[i].ParamName;
      if VarType(FParams.Param[i].Value) = varNull then
        Cells[1, i + 1] := ''
      else
        Cells[1, i + 1] := FParams.Param[i].Value;
    end;
  end;
end;

function TfmRepParam.OpenForm(Reporter: TCustomRepEngine): Integer;
begin
  FReporter := Reporter;
  AdjustSGSizes(sgData);
  AnalyzeParams;
  FillParams;
  if FReporter.Params.ParamCount > 0 then
  begin
    Result := ShowModal;
    if Result = mrOk then
      FReporter.Params.Assign(FParams)
    else
      FReporter.Params.Assign(FOldParams)
  end
  else
    Result := mrOk;
end;

procedure TfmRepParam.sgDataSetEditText(Sender: TObject; ACol, ARow: Integer;
  const Value: String);
begin
  if (FParams.ParamCount <= ARow-1)  or (ARow = 0) then exit;
  FParams.Param[ARow-1].Value := Value;
end;

procedure TfmRepParam.sgDataGetEditMask(Sender: TObject; ACol, ARow: Integer;
  var Value: String);
var
  S: String;
begin
  if (FParams = nil) or (FParams.ParamCount <= ARow-1) or (ARow = 0) then exit;

  S := FParams.Param[ARow-1].InputMask;
  if S <> EmptyStr then
    Value := S;
end;

procedure TfmRepParam.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_RETURN then btnOk.Click;
  if Key = VK_ESCAPE then btnCancel.Click;
end;

end.
