{**
*
*         Script Reporter Component library
*         Registration module
*          Vlad Ilushchenko 1998-99
*
*
**}
unit ScriptRep;

interface

uses Reporter;

type
  TScriptRep = class(TCustomScriptRep)
    property Prepared;
    property Processed;
    property ScriptFile;
    property OutputFile;
    property PageHeight;
    property OracleSession;
    property Params;
    property AppendFile;
    property MultiThreaded;
    property MaxPages;
    property OnTerminate;
    property OnNextPage;
    property OnStart;
    property AlignPages;
    property BeforeSubReportParsed;
    property OnLineProcessed;
  end;

procedure Register;

implementation

uses Classes, DsgnIntf, SysUtils, RepParamForm, Dialogs, RepKrnl, DebuggerForm;

type
  TParamsPropertyEditor = class(TPropertyEditor)
  public
    function GetValue: string; override;
    function GetAttributes: TPropertyAttributes; override;
    procedure Edit; override;
  end;

  TFileNamePropertyEditor = class(TStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure Edit; override;
  end;

  TScriptEditor = class(TComponentEditor)
  public
    procedure About;
    procedure Edit; override;
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerbCount: Integer; override;
    function GetVerb(Index: Integer): string; override;
  end;

procedure EditParams(RepEngine: TCustomRepEngine);
var
  F: TfmRepParam;
begin
  F := TfmRepParam.Create(nil);
  try
    F.OpenForm(RepEngine);
  finally
    F.Free;
  end;

end;

function TParamsPropertyEditor.GetValue: String;
begin
  Result := Format('%s', [TRepParams.ClassName]);
end;

function TParamsPropertyEditor.GetAttributes: TPropertyAttributes;
begin
  Result := [paReadOnly, paDialog];
end;

procedure TParamsPropertyEditor.Edit;
begin
  TScriptRep(GetComponent(0)).GetParams;
  EditParams(TScriptRep(GetComponent(0)));
  Designer.Modified;
end;

function TFileNamePropertyEditor.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog];
end;

procedure TFileNamePropertyEditor.Edit;
var
  OpenDialog: TOpenDialog;
begin
  OpenDialog := TOpenDialog.Create(nil);
  try
    OpenDialog.Filter := 'Script Reporter files|*.rsf|All files|*.*';
    OpenDialog.Options := [];
    OpenDialog.FileName := TScriptRep(GetComponent(0)).ScriptFile;
    if OpenDialog.Execute then
      TScriptRep(GetComponent(0)).ScriptFile := OpenDialog.FileName;

  finally
    OpenDialog.Free;
  end;
end;

{TScriptEditor}

procedure TScriptEditor.Edit;
begin
  with TScriptRep(Component) do
    OpenCompiler;
end;

procedure TScriptEditor.ExecuteVerb(Index: Integer);
begin
  case Index of
    0: About;
    1: Edit;
  end;
end;

function TScriptEditor.GetVerbCount: Integer;
begin
  Result := 2;
end;

function TScriptEditor.GetVerb(Index: Integer): string;
begin
  case Index of
    0: Result := 'About...';
    1: Result := 'Debugger...';
  end;
end;

procedure TScriptEditor.About;
begin
  with TScriptRep(Component).RepVersion do
    MessageDlg(Format('ScriptReporter v%d.%d.%d(%s). This component is FREEWARE.'#10+
                      ' Vlad Ilushchenko, Pavel Kilovaty, 1998-99'#10+
                      'Direct Oracle Access  Allround Automations'#10+
                      'support@allroundautomations.nl'#10+
                      #10+
                      'Special thanks to:'#10+
                      '------------------'#10+
                      'Andy Shapoval'#10+
                      'Serge Mironov-Kopisov'#10+
                      'Michael Rogozhnikov'#10+
                      #10+
                      'Free troubleshouting at reporter@isd.8m.com', [Major, Minor, Release, Comment]),
               mtInformation, [mbOk], 0);
end;

procedure Register;
begin
  RegisterComponents('ISD', [TScriptRep]);
  RegisterPropertyEditor(TypeInfo(TRepParams), TScriptRep, 'Params',
    TParamsPropertyEditor);
  RegisterPropertyEditor(TypeInfo(String), TScriptRep, 'ScriptFile',
    TFileNamePropertyEditor);
  RegisterComponentEditor(TScriptRep, TScriptEditor);
end;

initialization

RegisterClasses([TScriptRep, TRepParams]);

end.
